from whad.dot15d4.stack.manager import Dot15d4Manager
from whad.dot15d4.stack.service import Dot15d4Service
from whad.dot15d4.address import Dot15d4Address
from whad.dot15d4.stack.mac.constants import MACAddressMode
from whad.rf4ce.stack.nwk.exceptions import NWKTimeoutException
from whad.rf4ce.stack.nwk.database import NWKIB
from whad.rf4ce.stack.nwk.pairing import PairingEntry
from whad.rf4ce.crypto import generate_random_value, xor, RF4CECryptoManager, RF4CEKeyDerivation
from whad.scapy.layers.rf4ce import RF4CE_Command_Hdr, RF4CE_Cmd_Discovery_Request, \
    RF4CE_Hdr, RF4CE_Cmd_Key_Seed, RF4CE_Cmd_Pair_Request, RF4CE_Cmd_Pair_Response, \
    RF4CE_Cmd_Discovery_Response, RF4CE_Cmd_Ping_Request, RF4CE_Cmd_Ping_Response, \
    RF4CE_Data_Hdr, RF4CE_Vendor_Hdr, RF4CE_Vendor_MSO_Get_Attribute_Response, \
    RF4CE_Vendor_MSO_Hdr, RF4CE_Vendor_MSO_Check_Validation_Response
from whad.common.stack import Layer, alias, source, state
from whad.rf4ce.stack.apl import APLManager

from random import randint
from time import time, sleep
from struct import pack

import logging

logger = logging.getLogger(__name__)

class NWKService(Dot15d4Service):
    """
    This class represents a NWK service, exposing a standardized API.
    """
    def __init__(self, manager, name=None):
        super().__init__(
            manager,
            name=name,
            timeout_exception_class=NWKTimeoutException
        )

class NWKDataService(NWKService):
    """
    NWK service forwarding data packets.
    """
    def __init__(self, manager):
        super().__init__(manager, name="nwk_data")



    @Dot15d4Service.request("NLDE-DATA")
    def data(self, nsdu, pairing_reference=None, profile_id=0, vendor_id=0x0000, tx_options=0):
        """
        Request transmitting NSDU according to the provided parameters.
        """
        broadcast = bool(tx_options & 1)
        ieee_dest_mode = bool((tx_options & (1 << 1)) >> 1)
        ack_mode = bool((tx_options & (1 << 2)) >> 2)
        security_mode = bool((tx_options & (1 << 3)) >> 3)
        channel_agility = bool((tx_options & (1 << 4)) >> 4)
        channel_normalization = bool((tx_options & (1 << 5)) >> 5)
        vendor_mode = bool((tx_options & (1 << 6)) >> 6)


        frame_counter = self.database.get("nwkFrameCounter")
        self.database.set("nwkFrameCounter", frame_counter + 1)

        if frame_counter == 0xffffffff:
            # frame counter expired, terminate without further processing
            return False

        if vendor_mode:

            npdu = (
                RF4CE_Hdr(
                    security_enabled = security_mode,
                    frame_counter = frame_counter,
                    reserved = 1,
                    protocol_version = 1
                ) /
                RF4CE_Vendor_Hdr(
                    profile_id = profile_id,
                    vendor_id = vendor_id
                ) /
                nsdu
            )
        else:
            npdu = (
                RF4CE_Hdr(
                    security_enabled = security_mode,
                    frame_counter = frame_counter,
                    reserved = 1,
                    protocol_version = 1
                ) /
                RF4CE_Data_Hdr(
                    profile_id = profile_id,
                    vendor_id = vendor_id
                ) /
                nsdu
            )


        pairing_table = self.database.get('nwkPairingTable')
        try:
            if pairing_reference is not None:
                pairing_entry = pairing_table[pairing_reference]
            else:
                pairing_entry = None
        except IndexError:
            pairing_entry = None

        if broadcast:
            destination_pan_id = 0xFFFF
            destination_address = 0xFFFF

        else:
            if pairing_entry is not None:
                destination_pan_id = pairing_entry.destination_pan_id
                destination_address = pairing_entry.destination_ieee_address if ieee_dest_mode else pairing_entry.destination_network_address
                source_address = pairing_entry.source_network_address
            else:
                # no entry
                return False

        if pairing_entry is not None:
            channel = pairing_entry.channel
        else:
            channel = self.manager.get_layer('phy').get_channel()

        if security_mode:
            if pairing_entry is not None and pairing_entry.link_key is not None:
                link_key = pairing_entry.link_key

                enc_npdu = RF4CECryptoManager(
                    key = pairing_entry.link_key
                ).encrypt(
                    npdu,
                    source=pack('<Q', self.manager.get_layer('mac').database.get('macExtendedAddress')),
                    destination=pack('<Q', pairing_entry.destination_ieee_address),
                    rf4ce_only=True
                )
                npdu = enc_npdu

            else:
                return False

        if ack_mode:
            ack = self.manager.get_layer('mac').get_service("data").data(
                npdu,
                destination_address_mode=MACAddressMode.EXTENDED if ieee_dest_mode else MACAddressMode.SHORT,
                source_address_mode=MACAddressMode.SHORT,
                destination_pan_id=destination_pan_id,
                destination_address=destination_address,
                pan_id_suppressed = False,
                wait_for_ack=True
            )
            return ack
        else:
            self.manager.get_layer('mac').get_service("data").data(
                npdu,
                destination_address_mode=MACAddressMode.EXTENDED if ieee_dest_mode else MACAddressMode.SHORT,
                source_address_mode=MACAddressMode.SHORT,
                destination_pan_id=destination_pan_id,
                destination_address=destination_address,
                pan_id_suppressed = False,
                wait_for_ack=False
            )
            return True

    def on_data_npdu(self, npdu, source_address, destination_address, source_pan_id, destination_pan_id, link_quality=255):
        """
        Callback processing NPDU from NWK manager.

        It forwards the NPDU to indicate_data method.
        """
        self.indicate_data(npdu, source_address, destination_address, source_pan_id, destination_pan_id, link_quality)

    @Dot15d4Service.indication("NLDE-DATA")
    def indicate_data(self, npdu, source_address, destination_address, source_pan_id, destination_pan_id, link_quality=255):
        """
        Indication transmitting NSDU to upper layer.
        """
        #npdu.show()
        if RF4CE_Data_Hdr in npdu:
            nsdu = npdu[RF4CE_Data_Hdr][1:]
        elif RF4CE_Vendor_Hdr in npdu:
            nsdu = npdu[RF4CE_Vendor_Hdr][1:]
        else:
            nsdu = npdu

        rx_flags = (
            int(destination_address == 0xffff) |
            (int(npdu.security_enabled) << 1) |
            (int(npdu.frame_type == 3) << 2)
        )

        return (
            npdu,
            {
                "pairing_reference" : npdu.pairing_entry_index,
                "profile_id" : npdu.profile_id,
                "vendor_id" : npdu.vendor_id if hasattr(npdu, "vendor_id") else None,
                "link_quality" : link_quality,
                "rx_flags" : rx_flags
            }
        )



class NWKManagementService(NWKService):
    """
    NWK service forwarding management packets.
    """
    def __init__(self, manager):
        super().__init__(manager, name="nwk_management")

    @Dot15d4Service.request("NLME-PAIR")
    def pair_request(self, destination_pan_id, destination_ieee_address, list_of_device_types=[1], list_of_profiles=[192], key_exchange_transfer_count=3):
        """
        Request allowing the NWK layer to transmit a pairing request.
        """
        node_capability =  self.database.get("nwkcNodeCapabilities")
        channel_normalization_capable = (node_capability >> 3) & 1
        security_capable = (node_capability >> 2) & 1
        power_source = "battery_source" if ((node_capability >> 1) & 1) == 0 else "alternating_current_source"
        node_type = "controller" if (node_capability & 1) == 0 else "target"

        pairing_reference = None
        # Get entry
        pairing_entry = None
        pairing_table = self.database.get("nwkPairingTable")
        for pairing_reference in range(len(pairing_table)):
            if (
                pairing_table[pairing_reference].destination_ieee_address == destination_ieee_address and
                pairing_table[pairing_reference].destination_pan_id == destination_pan_id
            ):
                pairing_entry = pairing_table[pairing_reference]
                break

        if pairing_entry is None:
            # Create a new Pairing Entry here
            pairing_entry = PairingEntry(
                source_network_address = self.manager.get_layer('mac').database.get("macShortAddress"),
                channel = self.manager.get_layer('phy').get_channel(),
                destination_ieee_address = destination_ieee_address,
                destination_pan_id = destination_pan_id,
                destination_network_address = None, # unknown at this point
                capabilities = node_capability,
                frame_counter = 0,
                link_key = key_exchange_transfer_count # unknown for now, use it to transfer key_exchange_transfer_count
            )
            pairing_table.append(pairing_entry)
            pairing_reference = len(pairing_table) - 1

        frame_counter = self.database.get("nwkFrameCounter")
        self.database.set("nwkFrameCounter", frame_counter + 1)


        # Build a discovery request
        discovery_request = (
            RF4CE_Hdr(
                protocol_version=1,
                security_enabled=0,
                frame_counter=frame_counter
            ) /
            RF4CE_Command_Hdr() /
            RF4CE_Cmd_Discovery_Request(
                channel_normalization_capable = channel_normalization_capable,
                security_capable = security_capable,
                power_source = power_source,
                node_type = node_type,
                vendor_identifier = self.database.get("nwkVendorIdentifier"),
                vendor_string = self.database.get("nwkVendorString"),
                number_of_supported_profiles = len(list_of_profiles),
                profile_identifier_list = list_of_profiles,
                number_of_supported_device_types = len(list_of_device_types),
                device_type_list = list_of_device_types,
                user_string_specificied = int(self.database.get("nwkUserString") is not None),
                user_string = self.database.get("nwkUserString"),
                requested_device_type = 0xFF
            )
        )

        for _ in range(2):
            discovery_request.frame_counter = frame_counter
            #self.manager.get_layer('phy').set_channel(channel)
            ack = self.manager.get_layer('mac').get_service("data").data(
                discovery_request,
                destination_address_mode=MACAddressMode.SHORT,
                source_address_mode=MACAddressMode.EXTENDED,
                destination_pan_id=0xFFFF,
                destination_address=0xFFFF,
                pan_id_suppressed = False,
                wait_for_ack=True
            )
            frame_counter = self.database.get("nwkFrameCounter")
            self.database.set("nwkFrameCounter", frame_counter + 1)


        frame_counter = self.database.get("nwkFrameCounter")
        self.database.set("nwkFrameCounter", frame_counter + 1)

        #  Build the pairing response
        pairing_request = (
            RF4CE_Hdr(
                reserved = 1,
                protocol_version = 1,
                security_enabled = 0,
                frame_counter = frame_counter
            ) /
            RF4CE_Command_Hdr() /
            RF4CE_Cmd_Pair_Request
            (
                nwk_addr = 0xFFFE,# self.manager.get_layer('mac').database.get('macShortAddress'),
                channel_normalization_capable = channel_normalization_capable,
                security_capable = security_capable,
                power_source = power_source,
                node_type = node_type,
                vendor_identifier = self.database.get("nwkVendorIdentifier"),
                vendor_string = self.database.get("nwkVendorString"),
                number_of_supported_profiles = len(list_of_profiles),
                profile_identifier_list = list_of_profiles,
                number_of_supported_device_types = len(list_of_device_types),
                device_type_list = list_of_device_types,
                user_string_specificied = int(self.database.get("nwkUserString") is not None),
                user_string = self.database.get("nwkUserString"),
                key_exchange_transfer_count = key_exchange_transfer_count
            )
        )

        ack = self.manager.get_layer('mac').get_service("data").data(
            pairing_request,
            destination_address_mode=MACAddressMode.EXTENDED,
            source_address_mode=MACAddressMode.EXTENDED,
            destination_pan_id=destination_pan_id,
            destination_address=destination_ieee_address,
            pan_id_suppressed = False,
            wait_for_ack=True
        )

        if not ack:
            return None

        pair_response = None
        start = time()
        while (time() - start) < 3 and pair_response is None:
            try:
                pair_response = self.wait_for_packet(lambda pkt:RF4CE_Cmd_Pair_Response in pkt, timeout=0.01)
            except NWKTimeoutException:
                pass

        if pair_response is None:
            return None

        self.manager.get_layer('mac').set_short_address(pair_response.allocated_nwk_addr)
        self.manager.get_layer('mac').database.set("macPanId", destination_pan_id)
        pair_response.show()

        key_seeds = []
        while len(key_seeds) != key_exchange_transfer_count+1:
            key_seed = None
            start = time()
            while (time() - start) < 3 and key_seed is None:
                try:
                    key_seed = self.wait_for_packet(lambda pkt:RF4CE_Cmd_Key_Seed in pkt, timeout=0.01)
                    key_seeds.append(key_seed.seed_data)

                except NWKTimeoutException:
                    pass

        derivation = RF4CEKeyDerivation(seeds = key_seeds, seeds_number = key_exchange_transfer_count)
        key = derivation.key

        pairing_entry.link_key = key
        pairing_entry.destination_network_address = pair_response.nwk_addr
        pairing_entry.mark_as_active()

        frame_counter = self.database.get("nwkFrameCounter")
        self.database.set("nwkFrameCounter", frame_counter + 1)
        unenc_ping_req = (
            RF4CE_Hdr(
                protocol_version = 1,
                security_enabled = 1, # !!
                frame_counter = frame_counter
            ) /
            RF4CE_Command_Hdr() /
            RF4CE_Cmd_Ping_Request
            (
                ping_options = 0,
                ping_payload = generate_random_value(32)
            )
        )

        source_ieee_address = self.manager.get_layer('mac').database.get("macExtendedAddress")

        source_address_formatted = pack('<Q', pairing_entry.destination_ieee_address)
        destination_address_formatted = pack('<Q', source_ieee_address)


        enc_ping_req = RF4CECryptoManager(
            key = key
        ).encrypt(
            unenc_ping_req,
            destination=source_address_formatted,
            source=destination_address_formatted,
            rf4ce_only=True
        )

        ack = self.manager.get_layer('mac').get_service("data").data(
            enc_ping_req, #enc_ping_resp, #unenc_ping_resp,
            destination_address_mode=MACAddressMode.EXTENDED,
            source_address_mode=MACAddressMode.EXTENDED,
            destination_pan_id=0xFFFF,
            destination_address=destination_ieee_address,
            pan_id_suppressed = False,
            wait_for_ack=True
        )

        ping_resp = self.wait_for_packet(lambda pkt:RF4CE_Cmd_Ping_Response in pkt, timeout=3)
        ping_resp.show()
        return pairing_reference

    @Dot15d4Service.response("NLME-PAIR")
    def pair_response(self, destination_address, accept=True, list_of_device_types=[9], list_of_profiles=[192], pairing_reference=None):
        """
        Reponse allowing to the NWK layer to transmit a pairing response.
        """
        node_capability =  self.database.get("nwkcNodeCapabilities")
        channel_normalization_capable = (node_capability >> 3) & 1
        security_capable = (node_capability >> 2) & 1
        power_source = "battery_source" if ((node_capability >> 1) & 1) == 0 else "alternating_current_source"
        node_type = "controller" if (node_capability & 1) == 0 else "target"

        # Get entry
        pairing_entry = None
        try:
            pairing_table = self.database.get("nwkPairingTable")
            pairing_entry = pairing_table[pairing_reference]

        except (IndexError, TypeError):
            # something went wrong, terminate
            return False

        if pairing_entry is None:
            return False


        # Allocate network address
        network_address = None
        if (pairing_entry.capabilities & 1) == 0: #  controller
            already_allocated_addresses = []
            for node in self.database.get("nwkPairingTable"):
                already_allocated_addresses.append(node.source_network_address) # theoretically, it's always our own address
                already_allocated_addresses.append(node.destination_network_address)

            # remove duplicates
            already_allocated_addresses = list(set(already_allocated_addresses))
            while network_address is None or network_address in already_allocated_addresses:
                network_address =  randint(0x0000, 0xFFFD)

        else:
            network_address = 0xFFFE


        # Update network address of the pairing entry
        pairing_entry.destination_network_address = network_address # newly allocated address

        frame_counter = self.database.get("nwkFrameCounter")
        self.database.set("nwkFrameCounter", frame_counter + 1)

        #  Build the pairing response
        pairing_response = (
            RF4CE_Hdr(
                protocol_version = 1,
                security_enabled = 0,
                frame_counter = frame_counter
            ) /
            RF4CE_Command_Hdr() /
            RF4CE_Cmd_Pair_Response
            (
                status = (0 if accept else 1),
                allocated_nwk_addr = network_address,
                nwk_addr = self.manager.get_layer('mac').database.get('macShortAddress'),
                channel_normalization_capable = channel_normalization_capable,
                security_capable = security_capable,
                power_source = power_source,
                node_type = node_type,
                vendor_identifier = self.database.get("nwkVendorIdentifier"),
                vendor_string = self.database.get("nwkVendorString"),
                number_of_supported_profiles = len(list_of_profiles),
                profile_identifier_list = list_of_profiles,
                number_of_supported_device_types = len(list_of_device_types),
                device_type_list = list_of_device_types,
                user_string_specificied = int(self.database.get("nwkUserString") is not None),
                user_string = self.database.get("nwkUserString"),


            )
        )
        self.manager.get_layer('mac').database.set("macImplicitBroadcast", True)

        ack = self.manager.get_layer('mac').get_service("data").data(
            pairing_response,
            destination_address_mode=MACAddressMode.EXTENDED,
            source_address_mode=MACAddressMode.EXTENDED,
            destination_pan_id=0xFFFF,
            destination_address=destination_address,
            pan_id_suppressed = False,
            wait_for_ack=True
        )
        if not ack or not accept:
            pairing_table = self.database.get("nwkPairingTable")
            del pairing_table[pairing_reference]
            self.database.set("nwkPairingTable", pairing_table)
            return False
        else:
            if ((pairing_entry.capabilities >> 2) & 1) == 1 and security_capable == 1:

                # security enabled
                # Step 1: generate random 128 bit link key
                link_key = generate_random_value(128)

                # Step 2: generate key seeds
                key_exchange_transfer_count = pairing_entry.link_key

                seeds = [generate_random_value(640) for _ in range(key_exchange_transfer_count)]
                seeds_ph2 = [generate_random_value(128) for _ in range(4)]

                value = link_key
                for seed_ph2 in seeds_ph2:
                    value = xor(seed_ph2, value)

                seeds_ph2.append(value)

                value = b"".join(seeds_ph2)
                for seed in seeds:
                    value = xor(seed, value)

                seeds.append(value)

                for n in range(key_exchange_transfer_count + 1):

                    frame_counter = self.database.get("nwkFrameCounter")
                    self.database.set("nwkFrameCounter", frame_counter + 1)

                    key_seed = (
                        RF4CE_Hdr(
                            protocol_version = 1,
                            security_enabled = 0,
                            frame_counter = frame_counter
                        ) /
                        RF4CE_Command_Hdr() /
                        RF4CE_Cmd_Key_Seed
                        (
                            key_sequence_number = n,
                            seed_data = seeds[n]
                        )
                    )
                    ack = self.manager.get_layer('mac').get_service("data").data(
                        key_seed,
                        destination_address_mode=MACAddressMode.EXTENDED,
                        source_address_mode=MACAddressMode.EXTENDED,
                        destination_pan_id=0xFFFF,
                        destination_address=destination_address,
                        pan_id_suppressed = False,
                        wait_for_ack=True
                    )
                    if not ack:
                        pairing_table = self.database.get("nwkPairingTable")
                        del pairing_table[pairing_reference]
                        self.database.set("nwkPairingTable", pairing_table)
                        return False

                pairing_entry.link_key = link_key
                pairing_entry.mark_as_active()

                ping_req = None
                start = time()
                while (time() - start) < 3 and ping_req is None:
                    try:
                        ping_req = self.wait_for_packet(lambda pkt:RF4CE_Cmd_Ping_Request in pkt, timeout=0.01)
                    except NWKTimeoutException:
                        pass

                if ping_req is None or ping_req.ping_options != 0 or len(ping_req.ping_payload) != 4:
                    pairing_table = self.database.get("nwkPairingTable")
                    del pairing_table[pairing_reference]
                    self.database.set("nwkPairingTable", pairing_table)
                    return False

                frame_counter = self.database.get("nwkFrameCounter")
                self.database.set("nwkFrameCounter", frame_counter + 1)
                unenc_ping_resp = (
                    RF4CE_Hdr(
                        protocol_version = 1,
                        security_enabled = 0, # !!
                        frame_counter = frame_counter
                    ) /
                    RF4CE_Command_Hdr() /
                    RF4CE_Cmd_Ping_Response
                    (
                        ping_options = 0,
                        ping_payload = ping_req.ping_payload
                    )
                )

                ping_resp = (
                    RF4CE_Hdr(
                        protocol_version = 1,
                        security_enabled = 1,
                        frame_counter = frame_counter
                    ) /
                    RF4CE_Command_Hdr() /
                    RF4CE_Cmd_Ping_Response
                    (
                        ping_options = 0,
                        ping_payload = ping_req.ping_payload
                    )
                )
                source_ieee_address = self.manager.get_layer('mac').database.get("macExtendedAddress")

                source_address_formatted = pack('<Q', pairing_entry.destination_ieee_address)
                destination_address_formatted = pack('<Q', source_ieee_address)


                enc_ping_resp = RF4CECryptoManager(
                    key = pairing_entry.link_key
                ).encrypt(
                    ping_resp,
                    destination=source_address_formatted,
                    source=destination_address_formatted,
                    rf4ce_only=True
                )

                ack = self.manager.get_layer('mac').get_service("data").data(
                    unenc_ping_resp, #enc_ping_resp, #unenc_ping_resp,
                    destination_address_mode=MACAddressMode.EXTENDED,
                    source_address_mode=MACAddressMode.EXTENDED,
                    destination_pan_id=0xFFFF,
                    destination_address=destination_address,
                    pan_id_suppressed = False,
                    wait_for_ack=True
                )
                if ack:
                    pairing_entry.mark_as_active()
                return ack

            else:
                # no security
                pairing_entry.mark_as_active()
                return True

    @Dot15d4Service.indication("NLME-PAIR")
    def indicate_pair(self, pdu):
        """
        Indication signaling to the upper layer the reception of a pairing request.
        """
        source_pan_id = pdu.source_pan_id
        destination_pan_id = pdu.destination_pan_id
        source_address = pdu.source_address
        destination_address = pdu.destination_address

        node_capability = (
            (pdu.channel_normalization_capable << 3) |
            (pdu.security_capable << 2) |
            (pdu.power_source << 1) |
            pdu.node_type
        )

        vendor_identifier = pdu.vendor_identifier
        vendor_string = pdu.vendor_string

        application_capability = (
            ((pdu.number_of_supported_profiles & 0x02) << 4) |
            ((pdu.number_of_supported_device_types & 0x02) << 1) |
            pdu.user_string_specificied
        )

        user_string = pdu.user_string
        device_type_list = pdu.device_type_list
        profile_identifier_list = pdu.profile_identifier_list

        key_exchange_transfer_count = pdu.key_exchange_transfer_count

        pairing_entry = None
        for candidate in self.database.get("nwkPairingTable"):
            if (
                source_address == candidate.destination_address and
                destination_address == candidate.source_ieee_address and
                destination_pan_id == candidate.source_pan_id and
                node_capability == candidate.capabilities
            ):
                pairing_entry = candidate
                break

        # Note: maybe move this in a dedicated API ?
        pairing_table = self.database.get("nwkPairingTable")
        if pairing_entry is not None:
            pairing_reference = pairing_table.index(pairing_entry)
            # Do we update something here ? specification is unclear
        else:
            # Create a new Pairing Entry here
            pairing_entry = PairingEntry(
                source_network_address = self.manager.get_layer('mac').database.get("macShortAddress"),
                channel = self.manager.get_layer('phy').get_channel(),
                destination_ieee_address = source_address,
                destination_pan_id = source_pan_id,
                destination_network_address = None, # unknown at this point
                capabilities = node_capability,
                frame_counter = 0,
                link_key = key_exchange_transfer_count # unknown for now, use it to transfer key_exchange_transfer_count
            )
            pairing_table.append(pairing_entry)
            pairing_reference = len(pairing_table) - 1

        return (
            pairing_entry is None,
            {
                "source_pan_id":source_pan_id,
                "source_address":source_address,
                "node_capability":node_capability,
                "vendor_identifier":vendor_identifier,
                "vendor_string":vendor_string,
                "application_capability":application_capability,
                "user_string":user_string,
                "device_type_list":device_type_list,
                "profile_identifier_list":profile_identifier_list,
                "key_exchange_transfer_count":key_exchange_transfer_count,
                "pairing_reference":pairing_reference
            }
        )

    @Dot15d4Service.indication("NLME-DISCOVERY")
    def indicate_discovery(self, pdu):
        """
        Indication signaling to the upper layer the occurence of a discovery operation.
        """
        node_capability = (
            (pdu.channel_normalization_capable << 3) |
            (pdu.security_capable << 2) |
            (pdu.power_source << 1) |
            pdu.node_type
        )

        application_capability = (
            ((pdu.number_of_supported_profiles & 0x02) << 4) |
            ((pdu.number_of_supported_device_types & 0x02) << 1) |
            pdu.user_string_specificied
        )

        return (
            True,
            {
                "source_address" : pdu.source_address,
                "node_capability" : node_capability,
                "vendor_identifier" : pdu.vendor_identifier,
                "vendor_string" : pdu.vendor_string,
                "application_capability" : application_capability,
                "user_string":pdu.user_string if bool(pdu.user_string_specificied) else None,
                "device_type_list":pdu.device_type_list,
                "profile_identifier_list":pdu.profile_identifier_list,
                "requested_device_type":pdu.requested_device_type,
                "link_quality":pdu.link_quality
            }
        )

    @Dot15d4Service.indication("NLME-COMM-STATUS")
    def indicate_comm_status(self, status, pairing_reference=0xFF, pan_id=None, destination_address_mode=MACAddressMode.EXTENDED, destination_address=None):
        """
        Indication signaling to the upper layer the status of the communication.
        """
        return (status,
            {
                "pairing_reference":pairing_reference,
                "pan_id":pan_id,
                "destination_address_mode":destination_address_mode,
                "destination_address":destination_address
            }
        )

    @Dot15d4Service.response("NLME-DISCOVERY")
    def discovery_response(self, status, destination_address, list_of_device_types=[9], list_of_profiles=[192], link_quality=255):
        """
        Response allowing NWK layer to build and transmit a discovery response.
        """
        frame_counter = self.database.get("nwkFrameCounter")
        self.database.set("nwkFrameCounter", frame_counter + 1)

        node_capability =  self.database.get("nwkcNodeCapabilities")
        channel_normalization_capable = (node_capability >> 3) & 1
        security_capable = (node_capability >> 2) & 1
        power_source = "battery_source" if ((node_capability >> 1) & 1) == 0 else "alternating_current_source"
        node_type = "controller" if (node_capability & 1) == 0 else "target"


        # Build a discovery response
        discovery_response = (
            RF4CE_Hdr(
                protocol_version=1,
                security_enabled=0,
                frame_counter=frame_counter
            ) /
            RF4CE_Command_Hdr() /
            RF4CE_Cmd_Discovery_Response(
                status = int(status),
                channel_normalization_capable = channel_normalization_capable,
                security_capable = security_capable,
                power_source = power_source,
                node_type = node_type,
                vendor_identifier = self.database.get("nwkVendorIdentifier"),
                vendor_string = self.database.get("nwkVendorString"),
                number_of_supported_profiles = len(list_of_profiles),
                profile_identifier_list = list_of_profiles,
                number_of_supported_device_types = len(list_of_device_types),
                device_type_list = list_of_device_types,
                user_string_specificied = int(self.database.get("nwkUserString") is not None),
                user_string = self.database.get("nwkUserString"),
                discovery_req_lqi = link_quality
            )
        )
        ack = self.manager.get_layer('mac').get_service("data").data(
            discovery_response,
            destination_address_mode=MACAddressMode.EXTENDED,
            source_address_mode=MACAddressMode.EXTENDED,
            destination_pan_id=0xFFFF,
            destination_address=destination_address,
            wait_for_ack=True
        )

        self.indicate_comm_status(
            ack,
            0xFF,
            pan_id = 0xFFFF,
            destination_address_mode=MACAddressMode.EXTENDED,
            destination_address=destination_address
        )
        return ack


    @Dot15d4Service.request("NLME-AUTO-DISCOVERY")
    def auto_discovery(self, user_string_specificied=False, list_of_device_types=[9], list_of_profiles=[192], duration=60*50000*(10**6)): # ~ 15s
        """
        Request allowing NWK layer to auto respond to discovery requests.
        """
        duration_us = duration * self.manager.get_layer('phy').symbol_duration

        start = (time()*(10**6))
        match = 0
        while (time()*(10**6)) - start < duration_us:
            try:
                device_type_found = False
                profile_id_found = False

                discovery = self.wait_for_packet(lambda pkt:RF4CE_Cmd_Discovery_Request in pkt, timeout=0.01)

                if discovery.requested_device_type in list_of_device_types:
                    device_type_found = True

                for profile_id in discovery.profile_identifier_list:
                    if profile_id in list_of_profiles:
                        profile_id_found = True
                        break

                if device_type_found and profile_id_found:
                    match += 1
                else:
                    match = 0

                if match == 2:
                    frame_counter = self.database.get("nwkFrameCounter")
                    self.database.set("nwkFrameCounter", frame_counter + 1)

                    node_capability =  self.database.get("nwkcNodeCapabilities")
                    channel_normalization_capable = (node_capability >> 3) & 1
                    security_capable = (node_capability >> 2) & 1
                    power_source = "battery_source" if ((node_capability >> 1) & 1) == 0 else "alternating_current_source"
                    node_type = "controller" if (node_capability & 1) == 0 else "target"

                    # Build a discovery response
                    discovery_response = (
                        RF4CE_Hdr(
                            protocol_version=1,
                            security_enabled=0,
                            frame_counter=frame_counter
                        ) /
                        RF4CE_Command_Hdr() /
                        RF4CE_Cmd_Discovery_Response(
                            status = 0,
                            channel_normalization_capable = channel_normalization_capable,
                            security_capable = security_capable,
                            power_source = power_source,
                            node_type = node_type,
                            vendor_identifier = self.database.get("nwkVendorIdentifier"),
                            vendor_string = self.database.get("nwkVendorString"),
                            number_of_supported_profiles = len(list_of_profiles),
                            profile_identifier_list = list_of_profiles,
                            number_of_supported_device_types = len(list_of_device_types),
                            device_type_list = list_of_device_types,
                            user_string_specificied = int(self.database.get("nwkUserString") is not None),
                            user_string = self.database.get("nwkUserString"),
                            discovery_req_lqi = discovery.link_quality
                        )
                    )

                    ack = self.manager.get_layer('mac').get_service("data").data(
                        discovery_response,
                        destination_address_mode=MACAddressMode.EXTENDED,
                        source_address_mode=MACAddressMode.EXTENDED,
                        destination_pan_id=0xFFFF,
                        destination_address=discovery.source_address,
                        pan_id_suppressed = False,
                        wait_for_ack=True
                    )
                    if ack:
                        return (True, discovery.source_address)
            except NWKTimeoutException:
                pass

        return (False, None)

    @Dot15d4Service.request("NLME-DISCOVERY")
    def discovery(self, destination_address=0xFFFF, destination_pan_id=0xFFFF, list_of_device_types=[1], list_of_profiles=[192], search_device_type=0xFF, list_of_discovered_profiles=[], duration=60*50000*(10**6)):
        """
        Request allowing NWK layer to start a discovery operation.
        """
        node_descriptors = []

        frame_counter = self.database.get("nwkFrameCounter")

        node_capability =  self.database.get("nwkcNodeCapabilities")
        channel_normalization_capable = (node_capability >> 3) & 1
        security_capable = (node_capability >> 2) & 1
        power_source = "battery_source" if ((node_capability >> 1) & 1) == 0 else "alternating_current_source"
        node_type = "controller" if (node_capability & 1) == 0 else "target"

        # Build a discovery response
        discovery_request = (
            RF4CE_Hdr(
                protocol_version=1,
                security_enabled=0,
                frame_counter=frame_counter
            ) /
            RF4CE_Command_Hdr() /
            RF4CE_Cmd_Discovery_Request(
                channel_normalization_capable = channel_normalization_capable,
                security_capable = security_capable,
                power_source = power_source,
                node_type = node_type,
                vendor_identifier = self.database.get("nwkVendorIdentifier"),
                vendor_string = self.database.get("nwkVendorString"),
                number_of_supported_profiles = len(list_of_profiles),
                profile_identifier_list = list_of_profiles,
                number_of_supported_device_types = len(list_of_device_types),
                device_type_list = list_of_device_types,
                user_string_specificied = int(self.database.get("nwkUserString") is not None),
                user_string = self.database.get("nwkUserString"),
                requested_device_type = search_device_type
            )
        )
        for number_of_repetition in range(self.database.get("nwkMaxDiscoveryRepetitions")):
            for channel in [15, 20, 25]:
                frame_counter = self.database.get("nwkFrameCounter")
                self.database.set("nwkFrameCounter", frame_counter + 1)

                discovery_request.frame_counter = frame_counter
                self.manager.get_layer('phy').set_channel(channel)
                ack = self.manager.get_layer('mac').get_service("data").data(
                    discovery_request,
                    destination_address_mode=MACAddressMode.SHORT,
                    source_address_mode=MACAddressMode.EXTENDED,
                    destination_pan_id=destination_pan_id,
                    destination_address=destination_address,
                    pan_id_suppressed = False,
                    wait_for_ack=True
                )
                if ack:
                    try:
                        discovery = self.wait_for_packet(lambda pkt:RF4CE_Cmd_Discovery_Response in pkt, timeout=duration)
                        node_descriptors.append(discovery)
                    except NWKTimeoutException:
                        pass
                sleep(self.database.get("nwkDiscoveryRepetitionInterval") / (50000 * (10**6)))


        return node_descriptors

    def on_command_npdu(self, pdu, source_address, destination_address, source_pan_id, destination_pan_id, link_quality=255):
        pdu.link_quality = link_quality
        pdu.source_address = source_address
        pdu.destination_address = destination_address
        pdu.source_pan_id = source_pan_id
        pdu.destination_pan_id = destination_pan_id

        self.add_packet_to_queue(pdu)

        if pdu.command_identifier == 1:
            # If filtering is enabled and link quality is bad, drop
            lqi_threshold = self.manager.database.get("nwkDiscoveryLQIThreshold")
            if lqi_threshold not in (0, 0xFF) and link_quality < lqi_threshold:
                return
            self.indicate_discovery(pdu)

        elif pdu.command_identifier == 3:
            self.indicate_pair(pdu)

@state(NWKIB)
@alias('nwk')
class NWKManager(Dot15d4Manager):
    """
    This class implements the RF4CE Network manager (NWK).
    It handles network-level operations, such as discovery, association or network initiation.

    It exposes two services providing the appropriate API.
    """
    def init(self):
        self.add_service("data", NWKDataService(self))
        self.add_service("management", NWKManagementService(self))

    @source('mac', 'MCPS-DATA')
    def on_mcps_data(self, pdu, destination_pan_id, destination_address, source_pan_id, source_address, link_quality):
        """
        Callback processing MAC MCPS Data indication.
        """
        pdu.pairing_entry_index = None
        if pdu.security_enabled == 1:
            # if security is enable, check if we have enough info in pairing table
            pairing_table = self.database.get("nwkPairingTable")
            pairing_entry = None
            for candidate_pairing_entry in pairing_table:
                if (
                    (
                        candidate_pairing_entry.destination_ieee_address == source_address or
                        candidate_pairing_entry.destination_network_address == source_address
                    ) and
                    isinstance(candidate_pairing_entry.link_key, bytes) and
                    candidate_pairing_entry.is_active()
                ):
                    pairing_entry = candidate_pairing_entry
                    break

            if pairing_entry is not None:
                success = False
                source_ieee_address = self.get_layer('mac').database.get("macExtendedAddress")

                decryptor = RF4CECryptoManager(pairing_entry.link_key)

                source_address = pack('<Q', pairing_entry.destination_ieee_address)
                destination_address = pack('<Q', source_ieee_address)
                decrypted_packet, success = decryptor.decrypt(pdu, source=source_address, destination=destination_address, rf4ce_only=True)
                if success:
                    pdu = decrypted_packet
                    pdu.pairing_entry_index = pairing_table.index(pairing_entry)
                else:
                    return

        if pdu.frame_type == 1:
            self.get_service('data').on_data_npdu(pdu, source_address, destination_address, source_pan_id, destination_pan_id, link_quality)
        elif pdu.frame_type == 2:
            self.get_service('management').on_command_npdu(pdu, source_address, destination_address, source_pan_id, destination_pan_id, link_quality)
        elif pdu.frame_type == 3:
            self.get_service('data').on_data_npdu(pdu, source_address, destination_address, source_pan_id, destination_pan_id, link_quality)

NWKManager.add(APLManager)



from whad.dot15d4.stack.mac.constants import MACAddressMode

class ZCLClusterConfiguration:
    """
    This class defines the current configuration for a Zigbee Cluster Library cluster.
    """
    def __init__(   self,
                    destination_address_mode = None,
                    destination_address = None,
                    destination_endpoint = None,
                    transaction = None,
                    alias_address=None,
                    alias_sequence_number=0,
                    radius=30,
                    security_enabled_transmission=False,
                    use_network_key=True,
                    acknowledged_transmission=False,
                    fragmentation_permitted=False,
                    include_extended_nonce=False,
                    disable_default_response=False,
                    interpan=False,
                    asdu_handle=0,
                    source_address_mode=MACAddressMode.EXTENDED,
                    destination_pan_id=0xFFFF
    ):
        self.destination_address_mode = destination_address_mode
        self.destination_address = destination_address
        self.destination_endpoint = destination_endpoint
        self.transaction = transaction
        self.alias_address = alias_address
        self.alias_sequence_number = alias_sequence_number
        self.radius = radius
        self.security_enabled_transmission = security_enabled_transmission
        self.use_network_key = use_network_key
        self.acknowledged_transmission = acknowledged_transmission
        self.fragmentation_permitted = fragmentation_permitted
        self.include_extended_nonce = include_extended_nonce
        self.disable_default_response = disable_default_response
        self.interpan = interpan
        self.asdu_handle=asdu_handle
        self.source_address_mode=source_address_mode
        self.destination_pan_id=destination_pan_id



"""WHAD general sniffing tool

This utility implements a generic sniffer module, automatically adapted to every domain.
"""
import sys
import logging
from time import sleep
from argparse import ArgumentParser

from scapy.config import conf
from scapy.all import BrightTheme

from whad.cli.app import CommandLineDualOutputApp, run_app
from whad.cli.ui import error, warning, display_packet
from whad.exceptions import UnsupportedDomain, UnsupportedCapability
from whad.common.monitors import WiresharkMonitor, PcapWriterMonitor
from whad.device.unix import UnixSocketServer, UnixConnector
from whad.tools.utils import list_implemented_sniffers, get_sniffer_parameters, \
    build_configuration_from_args, gen_option_name
from whad.device import Bridge
from whad.phy.exceptions import UnsupportedFrequency

logger = logging.getLogger(__name__)


class WhadSniffOutputPipe(Bridge):
    """Whad sniff output pipe

    When wsniff is chained with another whad tool, it spawns a device
    based on the specified profile using the specified WHAD adapter and forward
    it to the chained tool. The chained tool will then forward packets back and forth.
    """

    def on_outbound(self, message):
        """Process outbound messages.

        Outbound packets are packets coming from our input connector,that need to be
        forwarded as packets to the next tool.
        """
        if hasattr(message, "to_packet") and hasattr(self.input, "process_packet"):
            pkt = message.to_packet()
            if pkt is not None:
                pkt = self.input.process_packet(pkt)
                msg = message.from_packet(pkt)
                super().on_outbound(msg)
        else:
            logger.debug('[wsniff][input-pipe] forward default outbound message %s', message)
            # Forward other messages
            super().on_outbound(message)

class WhadDomainSubParser(ArgumentParser):
    """
    Implements a Whad Domain subparser.
    """
    def warning(self, message):
        """Display a warning message in orange (if color is enabled)
        """
        warning(message)

    def error(self, message):
        """Display an error message in red (if color is enabled)
        """
        error(message)
        sys.exit(1)


class WhadSniffApp(CommandLineDualOutputApp):
    """Main WHAD sniffing application class.
    """

    def __init__(self, interface=True, description='WHAD generic sniffing tool',
                 pcap_argument=False):
        """Application uses an interface and has commands.
        """
        super().__init__(
            description=description,
            interface=interface,
            commands=False
        )

        self.add_argument(
            '--no-metadata',
            dest='metadata',
            action="store_false",
            help='Hide packets metadata'
        )
        self.add_argument(
            '--format',
            dest='format',
            action="store",
            default='repr',
            choices=['repr', 'show', 'raw', 'hexdump', 'tshark'],
            help='Indicate format to display packet'
        )

        self.add_argument(
            '-o',
            '--output',
            dest='output',
            default=None,
            type=str,
            help='Output PCAP file'
        )

        self.add_argument(
            '-w',
            '--wireshark',
            dest='wireshark',
            action='store_true',
            help='Enable wireshark monitoring'
        )

        if pcap_argument:
            self.add_argument(
                dest='pcap',
                help='PCAP file'
            )


        subparsers = self.add_subparsers(
            required=True,
            dest="domain",
            parser_class=WhadDomainSubParser,
            help='Domain in use'
        )

        self.build_subparsers(subparsers)



    def pre_run(self):
        """Pre-run operations: configure scapy theme.
        """
        super().pre_run()

        # If no color is not selected, configure scapy color theme
        if not self.args.nocolor:
            conf.color_theme = BrightTheme()

    def run(self):
        """Application main routine.
        """
        monitors = []
        sniffer = None
        # Launch pre-run tasks
        self.pre_run()

        try:

            # We need to have an interface specified
            if self.interface is not None:
                # We need to have a domain specified
                if self.args.domain is not None:
                    # Parse the arguments to populate a sniffer configuration
                    configuration = build_configuration_from_args(self.environment, self.args)

                    # Generate a sniffer based on the selected domain
                    sniffer = self.environment[self.args.domain]["sniffer_class"](self.interface)
                    # Add an event listener to display incoming events
                    # sniffer.add_event_listener(display_event)

                    sniffer.configuration = configuration
                    try:
                        pass
                    except Exception as e:
                        self.error("Error during configuration: " +repr(e))
                        raise KeyboardInterrupt from e
                    # If output parameter is selected, add a PCAP Writer monitor
                    if self.args.output is not None:
                        monitor_pcap = PcapWriterMonitor(self.args.output)
                        monitor_pcap.attach(sniffer)
                        monitor_pcap.start()
                        monitors.append(monitor_pcap)

                    # If wireshark parameter is selected, add a WiresharkMonitor
                    if self.args.wireshark:
                        monitor_wireshark = WiresharkMonitor()
                        monitor_wireshark.attach(sniffer)
                        monitor_wireshark.start()
                        monitors.append(monitor_wireshark)

                    sniffer.domain = self.args.domain
                    if self.is_stdout_piped():
                        # Create output proxy
                        #proxy = UnixSocketProxy(self.interface, params={"domain":self.args.domain})

                        # Create our unix socket server
                        unix_server = UnixConnector(UnixSocketServer(parameters={
                            'format': self.args.format,
                            'metadata':self.args.metadata,
                            'domain': self.args.domain
                        }))

                        # Wait for device to time out.
                        while not unix_server.device.opened:
                            if unix_server.device.timedout:
                                return
                            sleep(0.1)

                        # Create our packet bridge
                        logger.info("[wsniff] Starting our output pipe")
                        bridge = WhadSniffOutputPipe(sniffer, unix_server)
                        
                        # Start the sniffer
                        sniffer.start()

                        # Loop until the user hits CTL-C or interface disconnects
                        bridge.join()

                        logger.debug('wsniff: closing device')
                        unix_server.device.close()

                    else:
                        # Start the sniffer
                        sniffer.start()
                        # Iterates over the packet stream and display packets
                        for pkt in sniffer.sniff():
                            display_packet(
                                pkt,
                                show_metadata = self.args.metadata,
                                format = self.args.format
                            )
                else:
                    self.error("You need to specify a domain.")
            else:
                self.error("You need to specify an interface with option --interface.")

        except UnsupportedDomain:
            self.error(f"WHAD device doesn\'t support selected domain ({self.args.domain})")

        except UnsupportedCapability as unsupported_capability:
            self.error((f"WHAD device doesn't support selected capability"
                       f"({unsupported_capability.capability})"))
        except UnsupportedFrequency:
            self.error("WHAD interface doesn't support the requested frequency.")
        except KeyboardInterrupt:
            self.warning("sniffer stopped (CTRL-C)")
            sniffer.stop()
            sniffer.close()
            for monitor in monitors:
                monitor.close()
            sys.exit(1)

    def build_subparsers(self, subparsers):
        """
        Generate the subparsers argument according to the environment.
        """
        # List every domain implementing a sniffer
        self.environment = list_implemented_sniffers()

        # Iterate over domain, and get the associated sniffer parameters
        for name, domain in self.environment.items():
            domain["parameters"] = get_sniffer_parameters(
                domain["configuration_class"]
            )


            domain["subparser"] = subparsers.add_parser(
                name,
                description=f"WHAD {name.capitalize()} Sniffing tool"
            )

            # Iterate over every parameters, and add arguments to subparsers
            for (
                    parameter_name,
                    (parameter_type, parameter_default, parameter_base_class, parameter_help)
                ) in domain["parameters"].items():

                dest = parameter_name

                # If parameter is based on a dataclass, process a subparameter
                if parameter_base_class is not None:
                    _, parameter_name = parameter_name.split(".")

                parameter_name = gen_option_name(parameter_name)

                # Process parameter help and shortname
                if parameter_help is not None and "(" in parameter_help:
                    parameter_shortnames = [
                        f"-{i}" for i in
                        parameter_help.split("(")[1].replace(")","").split(",")
                    ]
                    parameter_help = parameter_help.split("(")[0]
                else:
                    parameter_shortnames = []


                # Process parameter type
                if parameter_type is not bool:
                    # If we got an int
                    if parameter_type is int:
                        # allow to provide hex arguments
                        parameter_type = lambda x: int(x,0)
                    # If we got a list, it is a list of string
                    if parameter_type is list:
                        parameter_default=[]
                        parameter_type=str
                        action = "append"
                    elif parameter_type is bytes:
                        parameter_type = bytes.fromhex
                        action = "store"
                    else:
                        action = "store"
                    # Add non-boolean argument to corresponding subparser
                    domain["subparser"].add_argument(
                        "--"+parameter_name,
                        *parameter_shortnames,
                        default=parameter_default,
                        action=action,
                        type=parameter_type,
                        dest=dest,
                        help=parameter_help
                    )
                else:
                    # Add boolean argument to corresponding subparser
                    domain["subparser"].add_argument(
                        "--"+parameter_name,
                        *parameter_shortnames,
                        action='store_true',
                        dest=dest,
                        help=parameter_help
                    )

def wsniff_main():
    """Main WHAD Sniffer routine.
    """
    app = WhadSniffApp()
    run_app(app)



"""WHAD Protocol PHY domain message abstraction layer.
"""
from typing import List
from dataclasses import dataclass, field, fields
from enum import IntEnum

from whad.protocol.phy.phy_pb2 import LoRaSpreadingFactor, LoRaCodingRate, JammingMode, \
    Endianness as PbEndianness, TXPower, Modulation as PbModulation
from whad.hub.registry import Registry
from whad.hub.message import HubMessage, pb_bind
from whad.hub import ProtocolHub
from whad.hub.metadata import Metadata
from whad.scapy.layers.phy import Phy_Packet_Hdr

class Commands:
    """PHY Commands
    """
    SetASKModulation = 0x00
    SetFSKModulation = 0x01
    SetGFSKModulation = 0x02
    SetBPSKModulation = 0x03
    SetQPSKModulation = 0x04
    Set4FSKModulation = 0x05
    SetMSKModulation  = 0x06
    GetSupportedFrequencies = 0x07
    SetFrequency = 0x08
    SetDataRate = 0x09
    SetEndianness = 0x0a
    SetTXPower = 0x0b
    SetPacketSize = 0x0c
    SetSyncWord = 0x0d
    Sniff = 0x0e
    Send = 0x0f
    SendRaw = 0x10
    Jam = 0x11
    Monitor = 0x12
    Start = 0x13
    Stop = 0x14
    SetLoRaModulation = 0x15
    ScheduleSend = 0x16

class Endianness(IntEnum):
    """PHY Endianness
    """
    BIG = PbEndianness.BIG
    LITTLE = PbEndianness.LITTLE

    def __str__(self):
        return self.name

class Modulation(IntEnum):
    """PHY Modulation
    """
    ASK = PbModulation.ASK
    FSK = PbModulation.FSK
    FOURFSK = PbModulation.FOURFSK
    GFSK = PbModulation.GFSK
    MSK = PbModulation.MSK
    BPSK = PbModulation.BPSK
    QPSK = PbModulation.QPSK
    LORA = PbModulation.LORA

    def __str__(self):
        return self.name

class Syncword(bytes):
    def __str__(self):
        return self.hex()

class SpreadingFactor:
    SF7 = LoRaSpreadingFactor.SF7
    SF8 = LoRaSpreadingFactor.SF8
    SF9 = LoRaSpreadingFactor.SF9
    SF10 = LoRaSpreadingFactor.SF10
    SF11 = LoRaSpreadingFactor.SF11
    SF12 = LoRaSpreadingFactor.SF12


class CodingRate:
    CR45 = LoRaCodingRate.CR45
    CR46 = LoRaCodingRate.CR46
    CR47 = LoRaCodingRate.CR47
    CR48 = LoRaCodingRate.CR48

class Jamming:
    CONTINUOUS = JammingMode.CONTINUOUS
    REACTIVE = JammingMode.REACTIVE

class TxPower:
    LOW = TXPower.LOW
    MEDIUM = TXPower.MEDIUM
    HIGH = TXPower.HIGH

@dataclass(repr=False)
class PhyMetadata(Metadata):
    frequency : int = None
    iq : list = field(default_factory=lambda: [])
    endianness : Endianness = None
    deviation : int = None
    datarate : int = None
    modulation : Modulation = None
    syncword : Syncword = None

    def convert_to_header(self):
        return Phy_Packet_Hdr(
            frequency=self.frequency,
            rssi=self.rssi,
            endianness=self.endianness,
            deviation=self.deviation,
            datarate=self.datarate,
            modulation=self.modulation,
            syncword=self.syncword,
            syncword_length=len(self.syncword)
        ), self.timestamp


    @classmethod
    def convert_from_header(cls, pkt):
        header = pkt[Phy_Packet_Hdr]

        return PhyMetadata(
            rssi = header.rssi,
            frequency = header.frequency,
            endianness = Endianness(header.endianness),
            deviation = header.deviation,
            datarate = header.datarate,
            modulation = Modulation(header.modulation),
            syncword = header.syncword,
            timestamp = int(100000 * pkt.time)
        )

def generate_phy_metadata(message):
    metadata = PhyMetadata()

    if message.rssi is not None:
        metadata.rssi = message.rssi

    metadata.frequency = message.frequency

    if message.timestamp is not None:
        metadata.timestamp = message.timestamp.sec*0.001 + message.timestamp.usec*0.000001


    if isinstance(message, ExtendedPacketReceived):
        metadata.endianness = message.endianness
        metadata.syncword = message.syncword
        metadata.deviation = message.deviation
        metadata.datarate = message.datarate
        metadata.modulation = message.modulation

    if isinstance(message, RawPacketReceived):
        if message.iq is not None:
            metadata.iq = [complex(message.iq[i], message.iq[i+1]) for i in range(0,len(message.iq)-1,2)]

    return metadata


@pb_bind(ProtocolHub, name="phy", version=1)
class PhyDomain(Registry):
    """WHAD PHY domain messages parser/factory.
    """

    NAME = 'phy'
    VERSIONS = {}

    def __init__(self, version: int):
        """Initializes a Dot15d4 domain instance
        """
        self.proto_version = version

    @staticmethod
    def parse(proto_version: int, message) -> HubMessage:
        """Parses a WHAD Dot15d4 Domain message as seen by protobuf
        """
        message_type = message.phy.WhichOneof('msg')
        message_clazz = PhyDomain.bound(message_type, proto_version)
        return message_clazz.parse(proto_version, message)

    def is_packet_compat(self, packet) -> bool:
        """Determine if a packet is a compatible BLE packet
        """
        return isinstance(packet.metadata, PhyMetadata)

    def convert_packet(self, packet) -> HubMessage:
        """Convert a Phy packet to SendPdu or SendBlePdu message.
        """
        if isinstance(packet.metadata, PhyMetadata):
            if packet.metadata.raw:
                return PhyDomain.bound('send_raw', self.proto_version).from_packet(
                    packet
                )
            else:
                return PhyDomain.bound('send', self.proto_version).from_packet(
                    packet
                )
        else:
            # Error
            return None

    def format(self, packet):
        """
        Converts a scapy packet with its metadata to a tuple containing a scapy packet with
        the appropriate header and the timestamp in microseconds.
        """
        header, timestamp =  packet.metadata.convert_to_header()
        return header / packet, timestamp

    def create_set_ask_mod(self, ook: bool = False) -> HubMessage:
        """Create a SetAskMod message

        :param ook: Enable On/Off Keying if set to True
        :type ook: bool
        :return: instance of `SetAskMod`
        """
        return PhyDomain.bound('mod_ask', self.proto_version)(
            ook=ook
        )

    def create_set_fsk_mod(self, deviation: int) -> HubMessage:
        """Create a SetFskMod message

        :param deviation: FSK deviation to use, in Hz
        :type deviation: int
        :return: instance of `SetFskMod`
        """
        return PhyDomain.bound('mod_fsk', self.proto_version)(
            deviation=deviation
        )

    def create_set_gfsk_mod(self, deviation: int) -> HubMessage:
        """Create a SetGfskMod message

        :param deviation: GFSK deviation to use, in Hz
        :type deviation: int
        :return: instance of `SetGfskMod`
        """
        return PhyDomain.bound('mod_gfsk', self.proto_version)(
            deviation=deviation
        )

    def create_set_bpsk_mod(self) -> HubMessage:
        """Create a SetBpskMod message
        :return: instance of `SetBpskMod`
        """
        return PhyDomain.bound('mod_bpsk', self.proto_version)()

    def create_set_qpsk_mod(self, offset: bool) -> HubMessage:
        """Create a SetQpskMod message

        :param offset: Use QPSK offset if set to True
        :type offset: bool
        :return: instance of `SetQpskMod`
        """
        return PhyDomain.bound('mod_qpsk', self.proto_version)(
            offset=offset
        )

    def create_set_4fsk_mod(self, deviation: int) -> HubMessage:
        """Create a Set4FskMod message

        :param deviation: 4FSK deviation to use, in Hz
        :type deviation: int
        :return: instance of `Set4FskMod`
        """
        return PhyDomain.bound('mod_4fsk', self.proto_version)(
            deviation=deviation
        )

    def create_set_msk_mod(self, deviation: int) -> HubMessage:
        """Create a SetMskMod message

        :param deviation: MSK deviation to use, in Hz
        :type deviation: int
        :return: instance of `SetMskMod`
        """
        return PhyDomain.bound('mod_msk', self.proto_version)(
            deviation=deviation
        )

    def create_set_lora_mod(self, bandwidth: int, sf: int, cr: int, preamble_length: int, \
                         enable_crc:bool = True, explicit_mode: bool = True, \
                         invert_iq: bool = False):
        """Create a SetLoRaMod message

        :param bandwidth: Bandwidth to use, in Hz
        :type bandwidth: int
        :param sf: LoRa spreading factor as defined in `SpreadingFactor`
        :type sf: int
        :param cr: LoRa coding rate as defined in `CodingRate`
        :type cr: int
        :param enable_crc: Include CRC in LoRa frames if set to True
        :type enable_crc: bool, optional
        :param explicit_mode: Enable LoRa explicit mode if set to True
        :type explicit_mode: bool, optional
        :param invert_iq: Invert IQ if set to True (used for downlink transmission)
        :type invert_iq: bool, optional
        :return: instance of `SetLoRaMod`
        """
        return PhyDomain.bound('mod_lora', self.proto_version)(
            bandwidth=bandwidth,
            sf=sf,
            cr=cr,
            preamble_length=preamble_length,
            enable_crc=enable_crc,
            explicit_mode=explicit_mode,
            invert_iq=invert_iq
        )


    def create_set_freq(self, frequency: int) -> HubMessage:
        """Create a SetFreq message

        :param frequency: Frequency to set
        :type frequency: int
        :return: instance of `SetFreq`
        """
        return PhyDomain.bound('set_freq', self.proto_version)(
            frequency=frequency
        )

    def create_get_supported_freqs(self) -> HubMessage:
        """Create a GetSupportedFreqs message

        :return: instance of `GetSupportedFreqs`
        """
        return PhyDomain.bound('get_supported_freq', self.proto_version)()

    def create_supported_freq_ranges(self, ranges: List[tuple]) -> HubMessage:
        """Create a SupportedFreqRanges message

        :param ranges: List of tuples containing range start and end frequencies
        :type ranges: list
        :return: instance of `SupportedFreqRanges`
        """
        # Create our supported freq ranges message
        msg: SupportedFreqRanges = PhyDomain.bound('supported_freq', self.proto_version)()

        # Add given ranges
        for start, end in ranges:
            msg.add(start, end)

        # Return created message
        return msg

    def create_sniff_mode(self, iq_stream: bool = False) -> HubMessage:
        """Create a SniffMode message

        :param iq_stream: Sniff I/Q stream if set to True, packets otherwise
        :type iq_stream: bool, optional
        :return: instance of `SniffMode`
        """
        return PhyDomain.bound('sniff', self.proto_version)(iq_stream=iq_stream)

    def create_jam_mode(self, mode: int) -> HubMessage:
        """Create a JamMode message

        :param mode: Jamming mode, as defined in `Jamming` class
        :type mode: int
        :return: instance of `JamMode`
        """
        return PhyDomain.bound('jam', self.proto_version)(mode=mode)

    def create_monitor_mode(self) -> HubMessage:
        """Create a MonitorMode message.

        :return: instance of `MonitorMode`
        """
        return PhyDomain.bound('monitor', self.proto_version)()

    def create_start(self) -> HubMessage:
        """Create a Start message

        :return: instance of `Start`
        """
        return PhyDomain.bound('start', self.proto_version)()

    def create_stop(self) -> HubMessage:
        """Create a Stop message

        :return: instance of `Stop`
        """
        return PhyDomain.bound('stop', self.proto_version)()

    def create_jammed(self, timestamp: int) -> HubMessage:
        """Create a Jammed notification

        :param timestamp: Timestamp at which the jamming has succeeded, in microseconds
        :type timestamp: int
        :return: instance of `Jammed`
        """
        return PhyDomain.bound('jammed', self.proto_version)(
            timestamp=timestamp
        )

    def create_montoring_report(self, timestamp: int, reports: List[int]) -> HubMessage:
        """Create a MonitoringReport notification message

        :param timestamp: Report timestamp
        :type timestamp: int
        :param reports: List of values contained in the monitoring report
        :type reports: list
        :return: instance of `MonitoringReport`
        """
        # Create our message
        msg = PhyDomain.bound('monitor_report', self.proto_version)(
            timestamp=timestamp
        )

        # Add reports
        for report_value in reports:
            msg.reports.append(report_value)

        # Return message
        return msg

    def create_set_datarate(self, datarate: int) -> HubMessage:
        """Create a SetDatarate message

        :param datarate: data rate to use
        :type datarate: int
        :return: instance of `SetDatarate`
        """
        return PhyDomain.bound('datarate', self.proto_version)(
            rate=datarate
        )

    def create_set_endianness(self, little: bool = True) -> HubMessage:
        """Create a SetEndianness message

        :param little: Use little-endian if set to True, big-endian otherwise
        :type little: bool
        :return: instance of `SetEndianness`
        """
        return PhyDomain.bound('endianness', self.proto_version)(
            endianness=Endianness.LITTLE if little else Endianness.BIG
        )

    def create_set_packet_size(self, size: int) -> HubMessage:
        """Create a SetPacketSize message

        :param size: Desired packet size
        :type size: int
        :return: instance of `SetPacketSize`
        """
        return PhyDomain.bound('packet_size', self.proto_version)(
            packet_size=size
        )

    def  create_set_tx_power(self, power: int) -> HubMessage:
        """Create a SetTxPower message

        :param power: Specify the TX power, as defined in `TxPower`
        :type power: int
        :return: instance of `SetTxPower`
        """
        return PhyDomain.bound('tx_power', self.proto_version)(
            power=power
        )

    def create_set_sync_word(self, syncword: bytes) -> HubMessage:
        """Create a SetSyncWord message

        :param syncword: Synchronization word to use
        :type syncword: bytes
        :return: instance of `SetSyncWord`
        """
        return PhyDomain.bound('sync_word', self.proto_version)(
            sync_word=syncword
        )

    def create_send_packet(self, packet: bytes) -> HubMessage:
        """Create a SendPacket message

        :param packet: Packet data to send
        :type packet: bytes
        :return: instance of `SendPacket`
        """
        return PhyDomain.bound('send', self.proto_version)(
            packet=packet
        )

    def create_send_raw_packet(self, iq: List[int]) -> HubMessage:
        """Create a SendPacket message

        :param iq: List of I/Q samples to send
        :type iq: list
        :return: instance of `SendRawPacket`
        """
        # Create our message
        msg = PhyDomain.bound('send_raw', self.proto_version)()

        # Add samples
        for sample in iq:
            msg.iq.append(sample)

        # Success
        return msg


    def create_packet_received(self, frequency: int, packet: bytes, rssi: int = None, \
                             timestamp: int = None, syncword : bytes = None, datarate : int = None, \
                             deviation : int = None, modulation : Modulation = None, \
                             endianness : Endianness = None) -> HubMessage:
        """Create a PacketReceived notification message

        :param frequency: Frequency on which the packet has been received
        :type frequency: int
        :param packet: Packet data
        :type packet: bytes
        :param rssi: Received signal strength indicator
        :type rssi: int, optional
        :param timestamp: Timestamp at which the packet has been received, in microseconds
        :type timestamp: int, optional
        :param syncword: Word used for radio synchronization
        :type syncword: bytes, optional
        :param datarate: Datarate
        :type datarate: int, optional
        :param deviation: Deviation if a frequency modulation is in use
        :type deviation: int, optional
        :param modulation: Modulation in use
        :type modulation: Modulation
        :param endianness: Endianness in use
        :type endianness: Endianness
        :return: instance of `PacketReceived`
        """

        msg = PhyDomain.bound('packet', self.proto_version)(
            frequency=frequency,
            packet=packet
        )
        # Add optional fields if provided
        if rssi is not None:
            msg.rssi = rssi
        if timestamp is not None:
            msg.timestamp = timestamp

        # Add modulation parameters
        if syncword is not None:
            msg.syncword = syncword
        if datarate is not None:
            msg.datarate = datarate
        if deviation is not None:
            msg.deviation = deviation
        if endianness is not None:
            msg.endian = endianness
        if modulation is not None:
            msg.modulation = modulation

        # Success
        return msg

    def create_raw_packet_received(self, frequency: int, packet: bytes, rssi: int = None, \
                             timestamp: int = None, iq: List[int] = None, \
                             syncword : bytes = None, datarate : int = None, \
                             deviation : int = None, modulation : Modulation = None, \
                             endianness : Endianness = None) -> HubMessage:
        """Create a RawPacketReceived notification message

        :param frequency: Frequency on which the packet has been received
        :type frequency: int
        :param packet: Packet data
        :type packet: bytes
        :param rssi: Received signal strength indicator
        :type rssi: int, optional
        :param timestamp: Timestamp at which the packet has been received, in microseconds
        :type timestamp: int, optional
        :return: instance of `PacketReceived`
        """
        msg = PhyDomain.bound('raw_packet', self.proto_version)(
            frequency=frequency,
            packet=packet
        )

        # Add optional fields if provided
        if rssi is not None:
            msg.rssi = rssi
        if timestamp is not None:
            msg.timestamp = timestamp
        if iq is not None:
            for sample in iq:
                msg.iq.append(sample)

        # Add modulation parameters
        if syncword is not None:
            msg.syncword = syncword
        if datarate is not None:
            msg.datarate = datarate
        if deviation is not None:
            msg.deviation = deviation
        if endianness is not None:
            msg.endian = endianness
        if modulation is not None:
            msg.modulation = modulation

        # Success
        return msg

    def create_schedule_packet(self, packet: bytes, timestamp: int) -> HubMessage:
        """Create a SchedulePacket message

        :param packet: Packet data
        :type packet: bytes
        :param timestamp: Timestamp at which the packet has to be sent, in microseconds
        :type timestamp: int
        :return: instance of `SchedulePacket`
        """
        # Create message
        msg = PhyDomain.bound('sched_send', self.proto_version)(
            packet=packet,
        )

        # Set timestamp value
        msg.timestamp = timestamp

        # Return message
        return msg

    def create_schedule_packet_response(self, packet_id: int, full: bool = False) -> HubMessage:
        """Create a SchedulePacketResponse message

        :param packet_id: Packet ID
        :type packet_id: int
        :param full: Set to True to indicate the schedule packet queue is full
        :type full: bool, optional
        :return: instance of `SchedulePacketResponse`
        """
        return PhyDomain.bound('sched_pkt_rsp', self.proto_version)(
            id=packet_id,
            full=full
        )

    def create_schedule_packet_sent(self, packet_id: int) -> HubMessage:
        """Create a SchedulePacketSent notification message

        :param packet_id: Packet id that has been sent
        :type packet_id: int
        :return: instance of SchedulePacketSent
        """
        return PhyDomain.bound('sched_pkt_sent', self.proto_version)(
            id=packet_id
        )

from .mod import SetAskMod, SetBpskMod, SetFskMod, SetGfskMod, SetLoRaMod, \
    SetMskMod, SetQpskMod, Set4FskMod
from .freq import GetSupportedFreqs, SetFreq, SupportedFreqRanges
from .packet import SetDatarate, SetEndianness, SetTxPower, SetPacketSize, SetSyncWord, \
    SendPacket, SendRawPacket, RawPacketReceived, PacketReceived, ExtendedPacketReceived
from .mode import SniffMode, JamMode, MonitorMode, Start, Stop, Jammed, MonitoringReport
from .schedule import SchedulePacket, ScheduledPacketSent, SchedulePacketResponse

__all__ = [
    'SetAskMod',
    'SetBpskMod',
    'SetFskMod',
    'SetGfskMod',
    'SetLoRaMod',
    'SetMskMod',
    'SetQpskMod',
    'Set4FskMod',
    'GetSupportedFreqs',
    'SetFreq',
    'SupportedFreqRanges',
    'SetDatarate',
    'SetEndianness',
    'SetTxPower',
    'SetPacketSize',
    'SetSyncWord',
    'SendPacket',
    'SendRawPacket',
    'PacketReceived',
    'ExtendedPacketReceived',
    'RawPacketReceived',
    'SniffMode',
    'JamMode',
    'Jammed',
    'MonitorMode',
    'MonitoringReport',
    'Start',
    'Stop',
    'SchedulePacket',
    'ScheduledPacketSent',
    'SchedulePacketResponse',
    'PhyDomain',
    'SpreadingFactor',
    'CodingRate',
    'Jamming',
    'TxPower'
]



from whad.ble.stack.constants import BT_MANUFACTURERS
from scapy.packet import Packet, bind_layers
from scapy.fields import ByteEnumField, ByteField, LEShortEnumField, LEShortField, \
    StrField, StrNullField, LELongField, XByteField
from scapy.layers.bluetooth import HCI_Command_Hdr, HCI_Event_Command_Complete, \
    HCI_Event_LE_Meta, LEMACField
from scapy.layers.bluetooth4LE import BTLEChanMapField

HCI_VERSIONS = LMP_VERSIONS = {
    0x00:   "1.0b",
    0x01:   "1.1",
    0x02:   "1.2",
    0x03:   "2.0",
    0x04:   "2.1",
    0x05:   "3.0",
    0x06:   "4.0",
    0x07:   "4.1",
    0x08:   "4.2",
    0x09:   "5.0",
    0x0a:   "5.1",
    0x0b:   "5.2",
    0x0c:   "5.3",
}

class HCI_Cmd_Read_Local_Version_Information(Packet):
    name = "Read Local Version Information Command"
    fields_desc = []


class HCI_Cmd_Complete_Read_Local_Version_Information(Packet):
    name = "Read Local Version Information Command Complete"
    fields_desc = [
        ByteEnumField("hci_version", None, HCI_VERSIONS),
        LEShortField("hci_subversion", None),
        ByteEnumField("lmp_version", None, LMP_VERSIONS),
        LEShortEnumField("company_identifier", None, BT_MANUFACTURERS),
        LEShortField("lmp_subversion", None),
    ]


class HCI_Cmd_Read_Local_Name(Packet):
    name = "Read Local Name"
    fields_desc = []


class HCI_Cmd_Complete_Read_Local_Name(Packet):
    name = "Read Local Name"
    fields_desc = [
                    StrNullField("local_name", None),
                    StrField("padding",None)
    ]

class HCI_Cmd_LE_Read_Supported_States(Packet):
    name = "LE Read Supported States Command"
    fields_desc = []


class HCI_Cmd_Complete_LE_Read_Supported_States(Packet):
    name = "LE Read Supported States Command Complete"
    fields_desc = [
        LELongField("supported_states", None)
    ]

class HCI_LE_Meta_Enhanced_Connection_Complete(Packet):
    name = "Enhanced Connection Complete"
    fields_desc = [
        ByteEnumField("status", 0, {0: "success"}),
        LEShortField("handle", 0),
        ByteEnumField("role", 0, {0: "master"}),
        ByteEnumField("patype", 0, {0: "public", 1: "random"}),
        LEMACField("paddr", None),
        LEMACField("localresolvprivaddr", None),
        LEMACField("peerresolvprivaddr", None),
        LEShortField("interval", 54),
        LEShortField("latency", 0),
        LEShortField("supervision", 42),
        XByteField("clock_latency", 5),
    ]


# Vendor specific commands - BD address modification
# Manufacturer : Texas Instruments (13)
class HCI_Cmd_TI_Write_BD_Address(Packet):
    name = "TI Write BD Address"
    fields_desc = [
        LEMACField("addr","\x00\x01\x02\x03\x04\x05")
    ]

# Manufacturer : Broadcom (15)
class HCI_Cmd_BCM_Write_BD_Address(Packet):
    name = "BCM Write BD Address"
    fields_desc = [
        LEMACField("addr","\x00\x01\x02\x03\x04\x05")
    ]

# Manufacturer : Zeevo (18)
class HCI_Cmd_Zeevo_Write_BD_Address(Packet):
    name = "Zeevo Write BD Address"
    fields_desc = [
        LEMACField("addr","\x00\x01\x02\x03\x04\x05")
    ]


# Manufacturer : Ericsson (0 / 57)
class HCI_Cmd_Ericsson_Write_BD_Address(Packet):
    name = "Ericsson Write BD Address"
    fields_desc = [
        LEMACField("addr","\x00\x01\x02\x03\x04\x05")
    ]

# Manufacturer : Cambridge Silicon Radios (10)
class HCI_Cmd_CSR_Write_BD_Address(Packet):
    name = "CSR Write BD Address"
    fields_desc = [
        LEMACField("addr","\x00\x01\x02\x03\x04\x05")
    ]

    def post_build(self,p,pay):
        payload = bytearray(b"\xc2\x02\x00\x0c\x00\x11G\x03p\x00\x00\x01\x00\x04\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00")

        payload[17] = p[2]
        payload[19] = p[0]
        payload[20] = p[1]
        payload[21] = p[3]
        payload[23] = p[4]
        payload[24] = p[5]

        return payload

class HCI_Cmd_CSR_Reset(Packet):
    name = "CSR Write BD Address"
    fields_desc = [
        StrField("bytes",b"\xc2\x02\x00\t\x00\x00\x00\x01@\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00")
    ]


# Manufacturer : ST (48)
class HCI_Cmd_ST_Write_BD_Address(Packet):
    name = "ST Write BD Address"
    fields_desc = [
        ByteField("user_id", 0xfe),
        ByteField("data_len",0x06),
        LEMACField("addr","\x00\x01\x02\x03\x04\x05"),
        StrField("padding","\x00"*247)
    ]


class HCI_Cmd_LE_Set_Host_Channel_Classification(Packet):
	name = "HCI Command LE Set Host Channel Classification"
	fields_desc = [
		BTLEChanMapField("chM" ,None)
	]


#bind_layers(HCI_Command_Hdr, HCI_Cmd_Read_Local_Version_Information,                        opcode=0x1001)
#bind_layers(HCI_Event_Command_Complete, HCI_Cmd_Complete_Read_Local_Version_Information,    opcode=0x1001)
#bind_layers(HCI_Command_Hdr,HCI_Cmd_Read_Local_Name,			                            opcode=0x0c14)
#bind_layers(HCI_Event_Command_Complete, HCI_Cmd_Complete_Read_Local_Name,                   opcode=0x0c14)
bind_layers(HCI_Command_Hdr, HCI_Cmd_LE_Read_Supported_States, ogf=0x08, ocf=0x001c)
bind_layers(HCI_Event_Command_Complete, HCI_Cmd_Complete_LE_Read_Supported_States,          opcode=0x201c)
bind_layers(HCI_Command_Hdr, HCI_Cmd_LE_Set_Host_Channel_Classification, ogf=0x08, ocf=0x0014)

bind_layers(HCI_Event_LE_Meta, HCI_LE_Meta_Enhanced_Connection_Complete,                    event=0xa)


bind_layers(HCI_Command_Hdr, HCI_Cmd_ST_Write_BD_Address,                                   opcode=0xfc22)
bind_layers(HCI_Command_Hdr, HCI_Cmd_Zeevo_Write_BD_Address,                                opcode=0xfc01)
bind_layers(HCI_Command_Hdr, HCI_Cmd_TI_Write_BD_Address,                                   opcode=0xfc06)
bind_layers(HCI_Command_Hdr, HCI_Cmd_Ericsson_Write_BD_Address,                             opcode=0xfc0d)
bind_layers(HCI_Command_Hdr, HCI_Cmd_BCM_Write_BD_Address,                                  opcode=0xfc01)
bind_layers(HCI_Command_Hdr, HCI_Cmd_CSR_Write_BD_Address,                                  opcode=0xfc00)
bind_layers(HCI_Command_Hdr, HCI_Cmd_CSR_Reset,                                             opcode=0xfc00)



"""Base RF4CE connector
"""
from whad.dot15d4.connector import Dot15d4

class RF4CE(Dot15d4):
    """
    RF4CE protocol connector.

    This connector drives a RF4CE-capable device with RF4CE-specific WHAD messages.
    It is required by various role classes to interact with a real device and pre-process
    domain-specific messages.
    """
    domain = "rf4ce"

    def __init__(self, device=None, synchronous=False):
        """
        Initialize the connector, open the device (if not already opened), discover
        the services (if not already discovered).
        """
        super().__init__(device, synchronous=synchronous, scapy_config='rf4ce')

    def sniff_rf4ce(self, channel:int = 15) -> bool:
        return super().sniff_dot15d4(channel=channel)



"""
LoRaWAN Gateway stack link-layer manager
"""
import logging

from time import sleep, time
from random import randint

from whad.common.stack import LayerState, Layer, alias, source, state,instance
from whad.scapy.layers.lorawan import MACPayloadUplink, PHYPayload, JoinAccept, JoinRequest, MACPayloadDownlink

from .mac import LWMacLayer
from ..crypto import MIC, derive_appskey, derive_nwkskey, encrypt_packet, decrypt_packet
from ..helpers import EUI

logger = logging.getLogger(__name__)

class LWGwLinkLayerState(LayerState):

    def __init__(self):
        super().__init__()

        # Keep track of connections
        self.connections = {}

        # Normal delays (1 second and then 2 seconds)
        self.rx_delay1 = 1.0
        self.rx_delay2 = 2.0

        # Join delays (5 seconds and then 6 seconds)
        self.join_delay1 = 5.0
        self.join_delay2 = 6.0

    def has_devaddr(self, dev_addr : int) -> bool:
        """Check if a device address is already registered

        :param dev_addr: Device address
        :type dev_addr: int

        :return: True if device address is known, False otherwise.
        :rtype: bool
        """
        return (dev_addr in self.connections)

    def register_connection(self, dev_addr : int, dev_eui : EUI = None, mac_node : str = None,
                            appskey : bytes = None, nwkskey : bytes = None, timestamp : float = None):
        """Register a device connection.

        Keeps track of MAC instance name associated with the device as well as the
        LoRaWAN 1.0 encryption keys.


        :param dev_addr: Device address
        :type dev_addr: int
        :param mac_node: MAC contextual layer instance name
        :type mac_node: str
        :param appskey: Application Session key
        :type appskey: bytes
        :param nwkskey: Network Session key
        :type nwkskey: bytes
        """
        if dev_addr not in self.connections:
            self.connections[dev_addr] = {
                'dev_eui': str(dev_eui),
                'mac_node': mac_node,
                'appskey': appskey,
                'nwkskey': nwkskey,
                'timestamp': timestamp
            }

    def update_connection(self, dev_addr : int, timestamp : float = None):
        """Update connection timestamp


        :param dev_addr: Device network address
        :type dev_addr: int
        :param timestamp: Timestamp in seconds
        :type timestamp: float
        """
        connection = self.get_connection(dev_addr)
        if connection is not None:
            if timestamp is not None:
                connection['timestamp'] = timestamp

    def get_connection(self, dev_addr : int) -> dict:
        """Retrieve a connection associated with a given
        device address.

        :param dev_addr: device address
        :type dev_addr: int
        :return: connection data
        :rtype: dict
        """
        if self.has_devaddr(dev_addr):
            return self.connections[dev_addr]
        else:
            return None


    def get_connection_from_node(self, mac_node : str) -> dict:
        """Get connection data from MAC layer node

        :param mac_node: MACLayer instance name
        :type mac_node: str
        
        :return: Connection data or None if not found
        :rtype: dict
        """
        for dev_addr in self.connections:
            if self.connections[dev_addr]['mac_node'] == mac_node:
                return self.connections[dev_addr]
        return None

@alias('ll')
@state(LWGwLinkLayerState)
class LWGwLinkLayer(Layer):
    """
    LoRaWAN Gateway link-layer manager.

    This layer handles the LoRaWAN over-the-air activation
    as well as encryption/decryption of incoming and outgoing
    communications.
    """

    def configure(self, options={}):
        """Configure Link layer state
        """
        # Process options
        if 'rx_delay1' in options:
            self.state.rx_delay1 = options['rx_delay1']
        if 'rx_delay2' in options:
            self.state.rx_delay2 = options['rx_delay2']
        if 'join_delay1' in options:
            self.state.join_delay1 = options['join_delay1']
        if 'join_delay2' in options:
            self.state.join_delay2 = options['join_delay2']

    def generate_devaddr(self) -> int:
        """Generate a free device network address

        :return: Attributed device network address
        :rtype: int
        """
        nok = True
        while nok:
            dev_addr = randint(0, 0xffffff)
            nok = self.state.has_devaddr(dev_addr)
        return dev_addr

    def on_join_request(self, frame : PHYPayload):
        """Process an incoming Join Request

        :param frame: LoRaWAN frame
        :type frame: PHYPayload
        """
        
        # Retrieve APPKey and APPEUI
        self.appkey = self.get_layer('phy').get_appkey()
        self.app_eui = self.get_layer('phy').get_appeui()

        logger.debug('processing a JoinRequest ...')
        # Check the Join Request against our APPKEY/APPEUI
        join_req = frame.getlayer(JoinRequest)
        exp_mic = MIC(self.appkey, bytes(frame)[:-4])
        if exp_mic == bytes(frame)[-4:]:
            logger.debug('JoinRequest MIC is valid')
            # Retrieve APP EUI amd DEV EUI from packet
            app_eui = EUI(join_req.join_eui)
            dev_eui = EUI(join_req.dev_eui)

            # Make sure APP EUI and DEV EUI are valid
            if app_eui == self.app_eui and self.get_layer('phy').is_device_allowed(dev_eui):

                # Device is allowed and application is known
                logger.debug(
                    "Device with EUI %s is allowed to join with app EUI %s",
                    dev_eui,
                    app_eui,
                )

                # Add node to our list of connections
                dev_addr = self.generate_devaddr()
                join_nonce = randint(0, 0xffffff)

                # Derive appskey and nwkskey
                dev_appskey = derive_appskey(self.appkey, join_nonce, 0x13, join_req.dev_nonce)
                dev_nwkskey = derive_nwkskey(self.appkey, join_nonce, 0x13, join_req.dev_nonce)

                # Instantiate a node for our connection
                conn_mac = self.instantiate(LWMacLayer)
                conn_mac.set_devaddr(dev_addr)

                # Save node information
                self.state.register_connection(
                    dev_addr,
                    dev_eui,
                    conn_mac.name,
                    dev_appskey,
                    dev_nwkskey,
                    time()
                )

                # Notify connector we got a connection
                self.get_layer('phy').on_device_joined(
                    dev_eui,
                    dev_addr,
                    dev_appskey,
                    dev_nwkskey
                )

                # Craft join accept                   
                join_accept = PHYPayload()/JoinAccept(
                    join_nonce = join_nonce,
                    home_netid=0x13,
                    dev_addr = dev_addr
                )

                # Encrypt response
                enc_ja = encrypt_packet(join_accept, appkey=self.appkey)

                # Send response
                ts = frame.metadata.timestamp/1000000.
                logger.debug("JoinRequest timestamp: %f", ts)
                logger.debug("will send JoinAccept at %f", ts + self.state.join_delay1)
                self.send('phy', enc_ja, timestamp=ts + self.state.join_delay1)
                sleep(self.state.join_delay1 + 0.5)
            else:
                if app_eui != self.app_eui:
                    logger.debug(
                        "Application %s is requested, expected application EUI %s",
                        app_eui,
                        self.app_eui,
                    )
                else:
                    logger.debug(
                        "Device with EUI %s is not allowed to access application EUI %s",
                        dev_eui,
                        app_eui,
                    )
        else:
            logger.debug(
                'JoinRequest MIC is wrong (got %s instead of %s)' % (
                bytes(frame)[-4:].hex(),
                exp_mic.hex()
            ))

    def add_provisioned_device(self, dev_addr : int, dev_eui : str, appskey : bytes,
                               nwkskey : bytes, upcount : int, dncount : int):
        """Add a pre-provisioned device to our current state.

        :param dev_addr: Device network address
        :type dev_addr: int
        :param dev_eui: Device EUI
        :type dev_eui: str
        :param dev_appskey: Device application session key
        :type dev_appskey: bytes
        :param dev_nwkskey: Device network encryption session key
        :type dev_nwkskey: bytes
        """
        logger.debug(
            "add a pre-provisioned device (address 0x%08x, eui: %s)",
            dev_addr,
            dev_eui,
        )

        # Instantiate a node for our connection
        conn_mac = self.instantiate(LWMacLayer)
        conn_mac.set_devaddr(dev_addr)
        logger.debug("created a LWMACLayer instance for device: %s", conn_mac.name)

        # Set uplink/downlink frame counters
        conn_mac.set_up_counter(upcount)
        conn_mac.set_down_counter(dncount)

        # Save node information
        self.state.register_connection(
            dev_addr,
            dev_eui,
            conn_mac.name,
            appskey,
            nwkskey,
        )

    def on_rejoin_request(self, rejoin_request):
        """Process a Rejoin Request

        Not implemented for now :(
        """
        pass

    def on_unconfirmed_data_up(self, unc_data_up : PHYPayload):
        """Process an unconfirmed data up.

        :param unc_data_up: LoRaWAN unconfirmed data up frame
        :type unc_data_up: PHYPayload
        """
        mac_layer = unc_data_up.getlayer(MACPayloadUplink)

        # Get the device address and check if it is known and that we have
        # at least seen one uplink packet
        connection = self.state.get_connection(mac_layer.dev_addr)
        if connection is not None and connection['timestamp'] is not None:
            # Update connection last packet timestamp
            ts = unc_data_up.metadata.timestamp/1000000.
            self.state.update_connection(
                mac_layer.dev_addr,
                ts
            )

            # Decrypt frame with the corresponding keys
            dec_unc_data_up = decrypt_packet(
                unc_data_up,
                appskey=connection['appskey'],
                nwkskey=connection['nwkskey']
            )

            # Forward the decrypted packet to MAC layer
            self.send(
                connection['mac_node'],
                dec_unc_data_up.getlayer(MACPayloadUplink),
                confirmed=False
            )
        else:
            logger.debug("Device address 0x%08x is not known", mac_layer.dev_addr)

    def on_confirmed_data_up(self, conf_data_up : PHYPayload):
        """Process a confirmed data up.

        Not supported yet.

        :param conf_data_up: LoRaWAN confirmed data up frame
        :type conf_data_up: PHYPayload
        """
        pass


    @source('phy')
    def on_phy_frame(self, frame: PHYPayload):
        """Process a frame coming from our gateway PHY layer.

        :param frame: LoRaWAN frame
        :type frame: PHYPayload
        """
        # Is it a join request ?
        if frame.mtype == 0:
            self.on_join_request(frame)
        # Or an unconfirmed data up ?
        elif frame.mtype == 2:
            self.on_unconfirmed_data_up(frame)
        # Or a confirmed data up ?
        elif frame.mtype == 4:
            self.on_confirmed_data_up(frame)
        # Or a rejoin request ?
        elif frame.mtype == 6:
            self.on_rejoin_request(frame)

    @instance('mac')
    def on_mac_response(self, inst_name : str, frame : MACPayloadDownlink, confirmed : bool = False):
        """Send downlink message to remote device.

        :param inst_name: Name of the MAC instance that called this method
        :type inst_name: str
        :param frame: LoRaWAN downlink MAC frame to send
        :type frame: MACPayloadDownlink
        :param confirmed: Confirmed downlink frame must be sent
        :type confirmed: bool
        """
        # Find node from instance name
        connection = self.state.get_connection_from_node(inst_name)
        if connection is not None:
            logger.debug(
                "[llm] MAC instance %s resolved to device eui %s",
                inst_name,
                connection['dev_eui'],
            )

            # Craft downlink frame
            packet = PHYPayload()/frame
            if confirmed:
                packet.mtype = 5
            else:
                packet.mtype = 3
            
            # Encrypt message for our target device           
            enc_pkt = encrypt_packet(
                packet,
                appskey=connection['appskey'],
                nwkskey=connection['nwkskey']
            )

            # Send response after rx1 delay
            self.send(
                'phy',
                enc_pkt,
                timestamp=connection['timestamp'] + self.state.rx_delay1
            )
            sleep(self.state.rx_delay1 + 0.5)
        else:
            logger.debug("[llm] MAC instance %s not found", inst_name)

    def on_data_received(self, dev_addr : int, data : bytes, upcount : int = 0):
        """Report data reception to phy

        :param dev_addr: Device network address
        :type dev_addr: int
        :param data: Data received
        :type data: bytes
        :param upcount: Uplink frame counter
        :type upcount: int
        """
        connection = self.state.get_connection(dev_addr)
        if connection is not None:
            # Connection found, report to PHY
            return self.get_layer('phy').on_data_received(
                connection['dev_eui'],
                dev_addr,
                data,
                upcount
            )
        else:
            logger.debug("[llm][on_data_received] Device address 0x%08x not found", dev_addr)
            return None



import logging
from typing import Union, Tuple

# packaging
from packaging.version import Version

# Scapy imports
from scapy.packet import Packet, Raw
from scapy.compat import raw
from scapy.config import conf
from scapy.layers.dot15d4 import Dot15d4 as Dot15d4NoFCS
from scapy.layers.dot15d4 import Dot15d4FCS

from whad.scapy.layers.dot15d4tap import Dot15d4TAP_Hdr
from whad.hub.dot15d4 import Dot15d4Metadata
# Main whad imports
from whad.scapy.layers.dot15d4tap import Dot15d4Raw
from whad.hub.discovery import Domain, Capability
from whad.cli.app import CommandLineApp
from whad.device.connector import Connector
from whad.helpers import message_filter, is_message_type
from whad.exceptions import UnsupportedDomain, UnsupportedCapability


# WHAD Protocol hub
from whad.hub.generic.cmdresult import Success, CommandResult
from whad.hub.dot15d4 import NodeAddress, Commands, NodeAddressType, PduReceived, \
    RawPduReceived, EnergyDetectionSample
from whad.hub.events import JammedEvt

logger = logging.getLogger(__name__)

class Dot15d4(Connector):
    """
    802.15.4 protocol connector.

    This connector drives a 802.15.4-capable device with 802.15.4-specific WHAD messages.
    It is required by various role classes to interact with a real device and pre-process
    domain-specific messages.
    """
    domain = "dot15d4"

    def __init__(self, device=None, synchronous=False, scapy_config='zigbee'):
        """
        Initialize the connector, open the device (if not already opened), discover
        the services (if not already discovered).
        """
        self.__ready = False
        super().__init__(device)

        # Capability cache
        self.__can_send = None
        self.__can_send_raw = None

        # Open device and make sure it is compatible
        self.device.open()
        self.device.discover()

        # Display a warning message if ButteRFly version is less than 1.0.2 as
        # a critical bug has been found and fixed in version 1.0.2. FCS values
        # will be wrong if using a version prior to 1.0.2.
        if device.info.fw_url == "https://github.com/whad-team/butterfly":
            if Version(device.info.version_str) < Version("1.0.2"):
                message = ((
                    "You are using a ButteRFly version prior to 1.0.2 that does not correctly compute FCS values, "
                    "this will result in invalid FCS values in packets and PCAP files that may cause errors when "
                    "used with other WHAD tools. Please consider upgrading firmware to the latest version "
                    "(see https://github.com/whad-team/butterfly). "
                    "You can also use `winstall --flash butterfly` to reprogram your USB dongle."
                ))

                # Use application warning method if available
                app = CommandLineApp.get_instance()
                if app is not None:
                    app.warning(message)
                else:
                    # If not available, use basic logging capabilities
                    logger.warning(message)

        # Check if device supports 802.15.4
        if not self.device.has_domain(Domain.Dot15d4):
            raise UnsupportedDomain("IEEE 802.15.4")
        else:
            self.__ready = True
            conf.dot15d4_protocol = scapy_config

        self.enable_synchronous(synchronous)

    def close(self):
        """
        Close the connector and the underlying device.
        """
        self.stop()
        self.device.close()

    def format(self, packet:Union[Dot15d4NoFCS,Dot15d4FCS]) -> Tuple[Dot15d4TAP_Hdr, int]:
        """
        Format a packet using the underlying translator.
        """
        if isinstance(packet, bytes):
            packet = Dot15d4FCS(packet)
        return self.hub.dot15d4.format(packet)

    def can_sniff(self) -> bool:
        """
        Determine if the device implements a sniffer mode.
        """
        commands = self.device.get_domain_commands(Domain.Dot15d4)
        return (
            (commands & (1 << Commands.Sniff)) > 0 and
            (commands & (1 << Commands.Start))>0 and
            (commands & (1 << Commands.Stop))>0
        )

    def can_set_node_address(self) -> bool:
        """
        Determine if the device can configure a Node address.
        """
        commands = self.device.get_domain_commands(Domain.Dot15d4)
        return (
            (commands & (1 << Commands.SetNodeAddress)) > 0
        )

    def can_be_end_device(self) -> bool:
        """
        Determine if the device implements an End Device role mode.
        """
        commands = self.device.get_domain_commands(Domain.Dot15d4)
        return (
            (commands & (1 << Commands.EndDeviceMode)) > 0 and
            (commands & (1 << Commands.Start))>0 and
            (commands & (1 << Commands.Stop))>0
        )

    def can_send(self) -> bool:
        """
        Determine if the device can transmit packets.
        """
        if self.__can_send is None:
            commands = self.device.get_domain_commands(Domain.Dot15d4)
            self.__can_send = ((commands & (1 << Commands.Send)) > 0 or (commands & (1 << Commands.SendRaw)) > 0)
        return self.__can_send

    def can_perform_ed_scan(self) -> bool:
        """
        Determine if the device can perform energy detection scan.
        """
        commands = self.device.get_domain_commands(Domain.Dot15d4)
        return (
            (commands & (1 << Commands.EnergyDetection)) > 0 and
            (commands & (1 << Commands.Start))>0 and
            (commands & (1 << Commands.Stop))>0
        )

    def support_raw_pdu(self) -> bool:
        """
        Determine if the device supports raw PDU.
        """
        if self.__can_send_raw is None:
            capabilities = self.device.get_domain_capability(Domain.Dot15d4)
            self.__can_send_raw = not (capabilities & Capability.NoRawData)
        return self.__can_send_raw

    def sniff_dot15d4(self, channel:int = 11) -> bool:
        """
        Sniff 802.15.4 packets (on a single channel).
        """
        if not self.can_sniff():
            raise UnsupportedCapability("Sniff")

        # Create a SniffMode message
        msg = self.hub.dot15d4.create_sniff_mode(channel)

        resp = self.send_command(msg, message_filter(CommandResult))
        return isinstance(resp, Success)

    def set_node_address(self, address:int, mode:NodeAddressType = NodeAddressType.SHORT) -> bool:
        """
        Modify 802.15.4 node address.
        """
        if not self.can_set_node_address():
            raise UnsupportedCapability("SetNodeAddress")

        # Create node address from parameters
        node_addr = NodeAddress(address, mode)

        # Create a SetNodAddress message
        msg = self.hub.dot15d4.create_set_node_address(node_addr)

        resp = self.send_command(msg, message_filter(CommandResult))
        return isinstance(resp, Success)

    def set_end_device_mode(self, channel:int = 11) -> bool:
        """
        Acts as a 802.15.4 End Device.
        """
        if not self.can_be_end_device():
            raise UnsupportedCapability("EndDevice")

        # Create EndDeviceMode message
        msg = self.hub.dot15d4.create_end_device_mode(channel)

        resp = self.send_command(msg, message_filter(CommandResult))
        return isinstance(resp, Success)


    def can_be_coordinator(self) -> bool:
        """
        Determine if the device implements a Coordinator role mode.
        """
        commands = self.device.get_domain_commands(Domain.Dot15d4)
        return (
            (commands & (1 << Commands.CoordinatorMode)) > 0 and
            (commands & (1 << Commands.Start))>0 and
            (commands & (1 << Commands.Stop))>0
        )


    def set_coordinator_mode(self, channel:int = 11) -> bool:
        """
        Acts as a 802.15.4 Coordinator.
        """
        if not self.can_be_coordinator():
            raise UnsupportedCapability("Coordinator")

        # Create EndDeviceMode message
        msg = self.hub.dot15d4.create_coord_mode(channel)

        resp = self.send_command(msg, message_filter(CommandResult))
        return isinstance(resp, Success)

    def send(self, pdu: Union[Packet, bytes, Dot15d4NoFCS, Dot15d4FCS], channel:int = 11) -> bool:
        """
        Send 802.15.4 packets (on a single channel).

        :param pdu: 802.15.4 packet to send
        :type pdu: scapy.layers.dot15d4.Dot15d4, scapy.layers.dot15d4.Dot15d4FCS
        :param channel: Channel on which the packet has to be sent
        :type channel: int
        :return: `True` if packet has been correctly sent, `False` otherwise.
        :rtype: bool
        """
        if self.can_send():
            metadata = Dot15d4Metadata()
            metadata.raw = False
            metadata.channel = channel

            # If PDU is provided as bytes, wrap it into a Raw packet
            if isinstance(pdu, bytes):
                pdu = Raw(pdu)

            if self.support_raw_pdu():
                metadata.raw = True

                if Dot15d4FCS not in pdu:
                    # Compute FCS if required by the hardware
                    packet = Dot15d4FCS(raw(pdu) + Dot15d4FCS().compute_fcs(raw(pdu)))
                else:
                    packet = pdu

            elif Dot15d4FCS in pdu:
                # Remove FCS if hardware cannot set it
                packet = Dot15d4NoFCS(raw(pdu)[:-2])
            else:
                packet = pdu

            if hasattr(packet, "reserved"):
                packet.reserved = packet.reserved

            # Set metadata
            packet.metadata = metadata
            # Send packet
            return super().send_packet(packet)
        else:
            return False


    def send_mac(self, pdu: bytes, channel: int = 11, add_fcs: bool = False):
        """
        Send raw 802.15.4 packets (on a single channel).

        :param pdu: 802.15.4 packet to send
        :type pdu: bytes
        :param channel: Channel on which the packet has to be sent
        :type channel: int
        :param add_fcs: Add FCS field if set to `True`
        :type add_fcs: bool
        :return: `True` if packet has been correctly sent, `False` otherwise.
        :rtype: bool
        """
        if self.can_send():
            # If raw mode is supported by the hardware, handle FCS value
            if self.support_raw_pdu():
                # Enable raw mode
                raw_mode = True

                # Add FCS if required
                if add_fcs:
                    fcs = Dot15d4FCS().compute_fcs(bytes(pdu))
                    packet = Dot15d4Raw(pdu + fcs)
                else:
                    packet = Dot15d4Raw(pdu)
            else:
                # Disable raw mode
                raw_mode = False

                # Cannot add/remove FCS, let hardware generate it
                logger.debug((
                    "[dot15d4::send_mac()] cannot add or remove FCS because HW"
                    "does not support raw packets, rollback to classic 802.15.4"
                    "frames with valid FCS."
                ))
                packet = Dot15d4Raw(pdu)

            # Add Dot15d4 metadata
            packet.metadata = Dot15d4Metadata()
            packet.metadata.raw = raw_mode
            packet.metadata.channel = channel

            # Send packet
            return super().send_packet(packet)

        # Failed at sending packet.
        return False

    def perform_ed_scan(self, channel:int = 11) -> bool:
        """
        Perform an Energy Detection scan.
        """
        if self.can_perform_ed_scan():
            # Create an EnergyDetectionMode message
            msg = self.hub.dot15d4.create_energy_detection_mode(channel)

            resp = self.send_command(msg, message_filter(CommandResult))
            return isinstance(resp, Success)
        else:
            return False

    def start(self) -> bool:
        """
        Start currently enabled mode.
        """
        # Create a Start message
        msg = self.hub.dot15d4.create_start()

        resp = self.send_command(msg, message_filter(CommandResult))
        return isinstance(resp, Success)

    def stop(self) -> bool:
        """
        Stop currently enabled mode.
        """
        # Create a Stop message
        msg = self.hub.dot15d4.create_stop()

        resp = self.send_command(msg, message_filter(CommandResult))
        return isinstance(resp, Success)

    def on_generic_msg(self, message):
        """
        Generic message handler.
        """

    def on_discovery_msg(self, message):
        """
        Discovery message handler.
        """

    def on_domain_msg(self, domain:str, message):
        """
        Domain message handler. Dispatches domain message to processing methods.
        """
        if not self.__ready:
            return

        assert domain == "dot15d4"
        if isinstance(message, EnergyDetectionSample):
            self.on_ed_sample(message.timestamp, message.sample)

    def on_packet(self, packet):
        """Dot15d4 packet dispatch.
        """
        if not self.__ready:
            return

        # Dispatch packet.
        if packet.metadata.raw:
            self.on_raw_pdu(packet)
        else:
            self.on_pdu(packet)

    def on_event(self, event):
        """Dot15d4 event dispatch.
        """
        if not self.__ready:
            return

        if isinstance(event, JammedEvt):
            self.on_jammed(event.timestamp)

    def on_raw_pdu(self, packet):
        """
        Raw PDU processing (Dot15d4FCS).
        """
        # Ugly hack but we need a forced rebuild in specific cases...
        if hasattr(packet, "reserved"):
            packet.reserved = packet.reserved

        pdu = Dot15d4NoFCS(packet.do_build()[:-2])
        pdu.metadata = packet.metadata
        self.on_pdu(pdu)

    def on_pdu(self, packet):
        """
        Normal PDU processing (Dot15d4).
        """
        # Enqueue PDU if in synchronous mode
        if self.is_synchronous():
            self.add_pending_packet(packet)
        else:
            pass

    def on_ed_sample(self, timestamp, sample):
        """
        Energy Detection sample processing.
        """
        pass

    def on_jammed(self, timestamp: int):
        """Jammed event handler.
        """
        pass

from whad.dot15d4.connector.sniffer import Sniffer
from whad.dot15d4.connector.enddevice import EndDevice
from whad.dot15d4.connector.coordinator import Coordinator

__all__ = [
    "Dot15d4",
    "Sniffer",
    "EndDevice",
    "Coordinator"
]



"""
ReadByType procedure

The ReadByType procedure is initiated by a ReadByType request, and
based on the required attribute start/end and type returns a
ReadByType response or an error.
"""
from typing import List
from struct import pack

from scapy.packet import Packet
from scapy.layers.bluetooth import ATT_Error_Response, ATT_Read_By_Type_Request, ATT_Read_By_Type_Response

from whad.ble.profile.attribute import UUID
from whad.ble.stack.att.constants import BleAttErrorCode, BleAttOpcode

from .attribute import find_attr_by_type
from .procedure import BleClientProcedure, BleServerProcedure, UnexpectedProcError

class ServerReadByTypeProcedure(BleServerProcedure):
    """ATT server ReadByType procedure for GATT server."""

    OPCODE = BleAttOpcode.READ_BY_TYPE_REQUEST

    def __init__(self, attributes: list, mtu: int):
        """Initialize our ServerReadByType procedure.

        :param attributes: List of GATT attributes
        :type attributes: list
        :param mtu: ATT MTU
        :type mtu: int
        """
        super().__init__(attributes, mtu)

    @classmethod
    def trigger(cls, request: Packet) -> bool:
        """Determine if the procedure should be triggered.

        :param request: Packet to check
        :type request: scapy.packet.Packet
        :rtype: bool
        :return: `True` if this procedure must be triggered, `False` otherwise.
        """
        return ATT_Read_By_Type_Request in request

    def process_request(self, request: Packet) -> List[Packet]:
        """React only on a ReadByType request.

        :param request: Request packet to process
        :type request: scapy.packet.Packet
        :rtype: scapy.packet.Packet, list
        :return: a single packet or a list of packets
        """
        if ATT_Read_By_Type_Request not in request:
            self.set_state(self.states.ERROR)
            raise UnexpectedProcError()

        # Extract ReadByType request
        request = request[ATT_Read_By_Type_Request]

        try:
            if request.start > request.end:
                self.set_state(self.states.DONE)
                return self.att_error_response(request.handle, BleAttErrorCode.INVALID_HANDLE)

            # List attributes by type
            attrs = find_attr_by_type(self.attributes, UUID(request.uuid),
                                      start_handle=request.start, end_handle=request.end)

            if len(attrs) == 0:
                self.set_state(self.states.DONE)
                return self.att_error_response(request.start, BleAttErrorCode.ATTRIBUTE_NOT_FOUND)

            # Build response value
            resp = b""
            attr_data_size = None
            for attr in attrs:
                # Append attribute info to our response
                attr_data = pack("<H", attr.handle) + attr.value

                # Save attribute data size if not set
                if attr_data_size is None:
                    attr_data_size = len(attr_data)

                # Append data if same size and not exceeding mtu
                if len(attr_data) == attr_data_size:
                    if self.mtu - len(resp) - 2 >= attr_data_size:
                        resp += attr_data
                    else:
                        break
                else:
                    break

            # Mark procedure as done
            self.set_state(self.states.DONE)

            # Return attribute list
            return [ ATT_Read_By_Type_Response(len=attr_data_size, handles=resp) ]


        except IndexError:
            # Attribute cannot be found
            self.set_state(self.states.DONE)
            return self.att_error_response(request.start, BleAttErrorCode.ATTRIBUTE_NOT_FOUND)

class ClientReadByTypeProcedure(BleClientProcedure):
    """ Client ReadByType procedure. """

    OPCODE = BleAttOpcode.READ_BY_TYPE_RESPONSE

    def __init__(self, start_handle: int, end_handle: int, type_uuid: UUID):
        """Initialize a client ReadByType procedure.

        :param start_handle: Start handle
        :type start_handle: int
        :param end_handle: End handle
        :type end_handle: int
        :param type_uuid: Attribute type UUID
        :type type_uuid: UUID
        """
        # Initialize parent procedure
        super().__init__([
            ATT_Read_By_Type_Request(
                start=start_handle,
                end=end_handle,
                uuid=type_uuid.value()
            )
        ])

    def process_request(self, request: Packet) -> List[Packet]:
        """Process incoming packet.

        :param request: Received packet
        :type request: Packet
        :return: List of packets to send in response
        :rtype: list
        """
        if ATT_Error_Response in request:
            self.set_result(request[ATT_Error_Response])
            self.set_state(self.states.ERROR)
        elif ATT_Read_By_Type_Response in request:
            # Save response as result
            self.set_result(request[ATT_Read_By_Type_Response])
            self.set_state(self.states.DONE)
        else:
            # Other packets: fail
            self.set_result(None)
            self.set_state(self.states.ERROR)

        # No other packets to send
        return []




"""This module provides a UnixSocket class that can be used with a WhadDeviceConnector
to interact with a unix socket connected to a Whad-enable device. This class implements
a Unix socket client that connects to a remote device through a Unix socket.

It also provides a dedicated connector to be used as a Unix socket server.

This class handles device connection, disconnection and read/write operations. All the
parsing magic is performed in our WhadDevice class.
"""

import sys
import os
import socket
import select
import re
import logging
from threading import Thread
from random import randint

from scapy.config import conf

from whad.exceptions import WhadDeviceNotReady, WhadDeviceDisconnected
from whad.hub.message import AbstractPacket

from .device import Device
from .connector import Connector

logger = logging.getLogger(__name__)

class UnixSocket(Device):
    """
    UnixSocket device class.
    """

    INTERFACE_NAME = "unix"

    @classmethod
    def list(cls):
        '''
        Returns a list of available Unix socket devices.

        To prevent identifying serial ports which are not compatible with WHAD, it implements
        a filtering mechanism based on vid, pid, manufacturer and / or product.
        '''
        devices = []

        try:
            with open("/proc/net/unix",'r', encoding="utf-8") as s:
                # Read /proc/net/unix (Linux only)
                proc_net_unix = s.read()

                # Extract all Unix sockets names, only keep those following the WHAD pattern:
                # whad_<random>.sock
                p = re.compile("^[0-9a-f]+: [0-9]+ [0-9]+ [0-9]+ [0-9]+ [0-9]+ [0-9]+ (.*)$",
                            re.I | re.M)
                for socket_ in p.findall(proc_net_unix):
                    _, filename = os.path.split(socket_)
                    if re.match('whad_[0-9a-f]+\\.sock', filename):
                        dev = UnixSocket(socket_)
                        devices.append(dev)
            return devices
        except IOError:
            # Not supported, cannot enumerate devices
            return devices

    def __init__(self, path=None):
        """
        Create device connection
        """
        super().__init__()

        # Connect to target Unix Socket device in non-blocking mode
        self.__path = path
        self.__fileno = None
        self.__socket = None
        self.__opened = False
        self.__stalled = False

    @property
    def identifier(self):
        '''
        Returns the identifier of the device (e.g., socket path).
        '''
        return self.__path


    def open(self):
        """
        Open device.
        """
        if not self.__opened:
            # Open Unix socket
            self.__socket = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
            self.__socket.connect(self.__path)
            self.__fileno = self.__socket.fileno()
            self.__opened = True

            # Ask parent class to run a background I/O thread
            super().open()


    def reset(self):
        """Reset device.

        This method is not supported by this type of device.
        """


    def close(self):
        """
        Close current device.
        """
        logger.debug('[UnixSocket] closing unix socket ...')
        # Close underlying device.
        if self.__socket is not None:
            self.__socket.close()

        # Clear fileno and status
        self.__fileno = None
        self.__opened = False

        # Ask parent class to stop I/O thread
        logger.debug('[UnixSocket] stopping I/O thread')
        super().close()


    def write(self, payload) -> int:
        """Writes data to the device. It relies on select() in order to make sure
        we are allowed to write to the device and wait without eating too much CPU
        if the device is not ready to be written to.

        :param bytes data: Data to write
        :returns: number of bytes written to the device
        """
        logger.debug("sending data to unix socket: %s", payload.hex())
        if not self.__opened:
            raise WhadDeviceNotReady()

        logger.debug("[%s] fileno: %s", self.interface, self.__fileno)
        nb_bytes_written = 0
        wlist = [self.__fileno]
        elist = [self.__fileno]
        _, writers, __ = select.select(
            [],
            wlist,
            elist
        )

        if len(writers) > 0:
            nb_bytes_written = self.__socket.send(payload)
        return nb_bytes_written

    def read(self) -> bytes:
        """Fetches data from the device, if there is any data to read. We call select()
        to make sure data is waiting to be read before reading it. Data is then sent to
        our parsing method through on_data_received() that will handle data reassembling
        and message parsing and dispatch.
        """
        try:
            # If not opened, device is not ready
            if not self.__opened:
                raise WhadDeviceNotReady()

            rlist = [self.__fileno]
            wlist = [self.__fileno]
            elist = [self.__fileno]

            readers, _, errors = select.select(
                rlist,
                wlist,
                elist,
                1.0
            )

            # Handle incoming messages if any
            #print(readers, errors)
            if len(readers) > 0:
                data = self.__socket.recv(1024)
                logger.info("[%s] Read data from socket: %s", self.interface, data)
                if len(data) > 0:
                    return data
                else:
                    logger.info("[%s] Read data, received empty buffer", self.interface)
                    # Unix socket client is stalled if we have pending messages
                    if not self.busy():
                        # If no pending message then consider the socket disconnected
                        logger.debug("[%s] Socket closed by remote peer.",
                                     self.interface)
                        raise WhadDeviceDisconnected()
                    elif not self.__stalled:
                        logger.debug((
                            "[%s] There are pending messages awaiting "
                            "for processing, consider unix socket stalled."),
                                     self.interface)
                        self.__stalled = True

            elif len(errors) > 0:
                logger.info("[%s] Unix socket in error", self.interface)
                raise WhadDeviceDisconnected()
        except ConnectionResetError:
            logger.error('Connection reset by peer')
        except Exception as err:
            raise WhadDeviceDisconnected() from err

    def change_transport_speed(self, speed):
        """Not supported by Unix socket devices.
        """

class UnixSocketServer(Device):
    """Unix socket server device
    """

    INTERFACE_NAME = "unix_server"

    def __init__(self, path: str = None, parameters: dict = None):
        """Create a WHAD unix socket.
        """
        super().__init__()

        # Indicate if a timeout occurred during opening
        self.__timedout = False

        # Create our Unix socket path
        if path is not None:
            # Use the provided Unix socket path
            self.__path = path
        else:
            # Generate socket path if not provided
            inst = randint(0x100000, 0x1000000)
            self.__path = f"/tmp/whad_{inst:x}.sock"

        logger.debug("Unix socket path: %s", self.__path)

        # Make sure this path is available
        try:
            if os.path.exists(self.__path) and os.path.isfile(self.__path):
                logger.debug("Unix socket path exists, deleting file...")
                os.unlink(self.__path)
            elif os.path.exists(self.__path):
                logger.debug("Unix socket path exist but is not a file")
                raise WhadDeviceNotReady
        except IOError as err:
            logger.debug("Error while cleaning Unix socket path %s", self.__path)
            raise WhadDeviceNotReady from err

        self.__socket = None
        self.__client = None
        self.__fileno = None
        self.__opened = False

        # Set unix socket server parameters
        if parameters is not None:
            self.__parameters = parameters
        else:
            self.__parameters = {}


    @property
    def identifier(self):
        '''
        Returns the identifier of the device (e.g., socket path).
        '''
        return self.__path


    @property
    def opened(self):
        '''
        Returns if a client is connected on the UNIX socket.
        '''
        return self.__opened

    @property
    def timedout(self):
        '''
        Returns if a client timed out.
        '''
        return self.__timedout


    def add_parameter(self, key: str, value):
        """Add a parameter to this Unix Socket connector.

        :param str key: parameter key
        :param object value: associated value
        """
        self.__parameters[key] = value

    def get_parameter(self, key : str):
        """Get value from a parameter.

        :param str key: parameter key
        """
        try:
            return self.__parameters[key]
        except KeyError:
            return None

    def get_url(self):
        """Return socket URL
        """
        if len(self.__parameters.keys()) == 0:
            return "unix://{self.__path}\n"

        params = '&'.join(['%s=%s' % item for item in self.__parameters.items()])
        return f"unix://{self.__path}?{params}\n"

    def open(self):
        """Open socket server and wait for a connection.
        """
        if not self.__opened:
            try:
                self.__socket = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
                self.__socket.settimeout(0.5)
                self.__socket.bind(self.__path)
                self.__socket.listen(5)

                logger.debug('Publishing unix socket info')
                sys.stdout.write(self.get_url())
                sys.stdout.flush()

                self.__client, infos = self.__socket.accept()
                logger.debug("unix socket server got a connection from %s", infos)
                self.__fileno = self.__client.fileno()
                logger.debug("fileno=%s", self.__fileno)
                self.__opened = True

                # Ask parent class to run a background I/O thread
                super().open()
            except BrokenPipeError:
                logger.error("Broken pipe.")
                self.__client = None
            except TimeoutError:
                logger.info("Timed out.")
                self.__timedout = True
                self.__client = None
            except Exception as other_err:
                logger.error("Another exception occurred: %s", other_err)

    def read(self) -> bytes:
        """Fetches data from the device, if there is any data to read. We call select()
        to make sure data is waiting to be read before reading it. Data is then sent to
        our parsing method through on_data_received() that will handle data reassembling
        and message parsing and dispatch.
        """
        try:
            if not self.__opened:
                raise WhadDeviceNotReady()

            rlist = [self.__fileno]
            wlist = []
            elist = [self.__fileno]

            readers, _, errors = select.select(
                rlist,
                wlist,
                elist,
                1
            )

            # Handle incoming messages if any
            if len(readers) > 0:
                data = self.__client.recv(1024)
                if len(data) > 0:
                    return data
                else:
                    logger.debug('No data received from client device')
                    raise WhadDeviceDisconnected()
            elif len(errors) > 0:
                raise WhadDeviceDisconnected()

        except ConnectionResetError:
            logger.debug('Connection reset by peer')
        except Exception as err:
            #logger.error('Unknown exception occurred (%s)' % err)
            raise WhadDeviceDisconnected() from err

    def write(self, payload):
        """Writes data to the device. It relies on select() in order to make sure
        we are allowed to write to the device and wait without eating too much CPU
        if the device is not ready to be written to.

        :param bytes payload: Data to write
        :returns: number of bytes written to the device
        """
        logger.debug("sending data to unix server client socket: %s", payload.hex())
        if not self.__opened:
            raise WhadDeviceNotReady()

        nb_bytes_written = 0
        wlist = [self.__fileno]
        elist = [self.__fileno]
        try:
            _, writers, errors = select.select(
                [],
                wlist,
                elist,
                .2
            )
        except TypeError:
            return 0

        if len(writers) > 0:
            nb_bytes_written = self.__client.send(payload)
        elif len(errors) > 0:
            raise WhadDeviceDisconnected()
        return nb_bytes_written

    def close(self):
        """Close connection.
        """
        if self.__client is not None:
            logger.debug("Closing UnixSocketServer client ...")
            self.__client.close()
            logger.debug("Client closed")
        self.__fileno = None

    def reset(self):
        """Reset device.

        This method is not supported by this type of device.
        """

    def change_transport_speed(self, speed):
        """Not supported by Unix socket devices.
        """

class UnixConnector(Connector):
    """Dummy connector for Unix socket.

    Connector is locked by default.
    """

    def __init__(self, device, locked: bool = True):
        """Create a locked connector.
        """
        if locked:
            # Create an empty connector, attach to no device
            super().__init__(None)

            # Lock connector
            self.lock()

            # Attach to device
            self.set_device(device)
            device.set_connector(self)
        else:
            #Create a dummy connector.
            super().__init__(device)

        # Open device if not already opened
        if not device.opened:
            # Unix socket server has already sent some messages but is now
            # closed, the previous tool must have finished earlier. We don't
            # need to open the socket again (it would fail) but simply consider
            # it open in order to process messages.
            if device.busy():
                logger.debug("[%s] Unix socket has sent messages and closed.",
                             device.interface)
            else:
                # Open socket if no pending message
                device.open()

    def on_discovery_msg(self, message):
        pass

    def on_generic_msg(self, message):
        pass

    def on_domain_msg(self, domain, message):
        pass

    def on_packet(self, packet):
        pass

    def on_event(self, event):
        pass

class UnixSocketConnector(Connector):
    """Unix socket server connector.

    This connector will create a Unix socket and accept a single connection
    to this socket. It will then forward every message receveived from the
    underlying WHAD device to the connected client, and the messages received
    from the client to the device (passthrough mode).

    WHAD Device <-> Unix socket (created)
    """

    def __init__(self, device, path=None):
        """Create a UnixSocketConnector
        """
        # No client connected
        self.__client = None
        self.__socket = None

        # Input pipes
        self.__inpipe = bytearray()

        # Parameters
        self.__parameters = {}

        # Shutdown require (to force exit in  ̀`serve()`)
        self.__shutdown_required = False

        # Initialize parent class (device connector)
        super().__init__(None)
        self.lock()
        self.set_device(device)
        device.set_connector(self)

        # Create our Unix socket path
        if path is not None:
            # Use the provided Unix socket path
            self.__path = path
        else:
            # Generate socket path if not provided
            inst = randint(0x100000, 0x1000000)
            self.__path = f"/tmp/whad_{inst:x}.sock"

        logger.debug("Unix socket path: %s", self.__path)

        # Make sure this path is available
        try:
            if os.path.exists(self.__path) and os.path.isfile(self.__path):
                logger.debug('Unix socket path exists, deleting file...')
                os.unlink(self.__path)
            elif os.path.exists(self.__path):
                logger.debug('Unix socket path exist but is not a file')
                raise WhadDeviceNotReady
        except IOError as err:
            logger.debug("Error while cleaning Unix socket path %s", self.__path)
            raise WhadDeviceNotReady from err

    def add_parameter(self, key: str, value):
        """Add a parameter to this Unix Socket connector.

        :param str key: parameter key
        :param object value: associated value
        """
        self.__parameters[key] = value

    def get_parameter(self, key : str):
        """Get value from a parameter.

        :param str key: parameter key
        """
        try:
            return self.__parameters[key]
        except KeyError:
            return None

    def send_message(self, message, keep=None):
        """Send message into the unix socket.
        """
        logger.debug("[%s] sending message %s", self.device.interface, message)
        self.device.send_message(message, keep=keep)

    def on_data_received(self, data):
        """Handle received data from the unix socket.
        """
        logger.debug("received raw data from socket: %s", data.hex())
        self.__inpipe.extend(data)
        while len(self.__inpipe) > 2:
            # Is the magic correct ?
            if self.__inpipe[0] == 0xAC and self.__inpipe[1] == 0xBE:
                # Have we received a complete message ?
                if len(self.__inpipe) > 4:
                    msg_size = self.__inpipe[2] | (self.__inpipe[3] << 8)
                    if len(self.__inpipe) >= (msg_size+4):
                        raw_message = self.__inpipe[4:4+msg_size]

                        # Parse message using our Protocol Hub
                        _msg = self.hub.parse(bytes(raw_message))

                        # Send to device
                        logger.debug(("WHAD message successfully parsed, "
                                     "forward to underlying device"))
                        self.send_message(_msg)

                        # Notify message
                        self.on_msg_sent(_msg)

                        # Chomp
                        self.__inpipe = self.__inpipe[msg_size + 4:]
                    else:
                        break
                else:
                    break
            else:
                # Nope, that's not a header
                while len(self.__inpipe) >= 2:
                    if (self.__inpipe[0] != 0xAC) or (self.__inpipe[1] != 0xBE):
                        self.__inpipe = self.__inpipe[1:]
                    else:
                        break

    def process_message(self, message):
        """Process received message.
        """
        # Send to device
        logger.debug(("WHAD message successfully parsed, "
                        "forward to underlying device"))
        self.send_message(message)

        # Notify message
        self.on_msg_sent(message)

    def shutdown(self):
        """Shutdown unix socket.
        """
        self.__shutdown_required = True

    def serve(self, timeout=None):
        """Serve Unix socket

        Forward data received through our Unix socket to the underlying device.

        :param float timeout: number of seconds to wait for incoming data
        """
        logger.debug('Creating Unix socket server')
        self.__socket = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
        self.__socket.settimeout(0.5)
        self.__socket.bind(self.__path)
        self.__socket.listen(5)

        logger.debug('Publishing unix socket info')
        sys.stdout.write(self.get_url())
        sys.stdout.flush()

        try:
            while not self.__shutdown_required:
                self.__client,_ = self.__socket.accept()
                self.unlock()
                try:
                    while not self.__shutdown_required:
                        rlist = [self.__client.fileno()]
                        readers, _, errors = select.select(rlist, [], rlist, timeout)

                        # Do we have pending data ?
                        if len(readers) > 0:
                            data = self.__client.recv(4096)
                            # Is socket closed ?
                            if len(data) == 0:
                                # Exit serve loop
                                break

                            # Process received data
                            self.on_data_received(data)
                        elif len(errors) > 0:
                            logger.debug('Error detected on server socket, exiting.')
                            break
                except ConnectionResetError:
                    logger.error("connection reset")

                self.__client = None

        except BrokenPipeError:
            logger.error("Broken pipe.")
            self.__client = None
        except TimeoutError:
            logger.debug("Timeout error.")
            self.__client = None
        except Exception as other_err:
            logger.debug("Unknown error of type %s: %s", type(other_err), other_err)

    # Message callbacks
    def on_any_msg(self, message):
        """Callback function to process incoming discovery messages.

        This method MUST be overriden by inherited classes.

        :param message: Discovery message
        """
        try:
            logger.debug("Received a message (%s) from device, forward to client if any (%s)", message, self.__client)
            if self.__client is not None:
                # Convert message into bytes
                raw_message = message.serialize()

                # Define header
                header = [
                    0xAC, 0xBE,
                    len(raw_message) & 0xff,
                    (len(raw_message) >> 8) & 0xff
                ]

                # Send header followed by serialized message
                self.__client.send(bytes(header))
                self.__client.send(raw_message)
                logger.debug('Message sent to client')
        except BrokenPipeError:
            logger.debug('Client socket disconnected')

    def on_msg_sent(self, message):
        """Outgoing message handler, actually doing nothing but needed by
        classes that inherit from WhadDeviceConnector.
        """

    def on_generic_msg(self, message):
        """Callback function to process incoming generic messages.

        This method MUST be overriden by inherited classes.

        :param message: Generic message
        """

    def on_discovery_msg(self, message):
        pass

    def on_domain_msg(self, domain, message):
        """Callback function to process incoming domain-related messages.

        This method MUST be overriden by inherited classes.

        :param message: Domain message
        """

    def on_packet(self, packet):
        """Callback function to process incoming packets.
        """

    def on_event(self, event):
        """Handle generic WHAD events.
        """

    def get_url(self):
        """Return socket URL
        """
        if len(self.__parameters.keys()) == 0:
            return f"unix://{self.__path}\n"

        params = '&'.join(["%s=%s" % item for item in self.__parameters.items()])
        return f"unix://{self.__path}?{params}\n"

class UnixSocketCallbacksConnector(UnixSocketConnector):
    """
    Unix Socket Callbacks Connector.

    Allows to configure callbacks to extract & alter the packet stream on the fly.
    """
    def __init__(self, device, path=None):
        super().__init__(device, path)
        conf.dot15d4_protocol = self.get_parameter('domain')

    def send_message(self, message, keep=None):
        """Send message to unix socket.
        """
        on_tx_packet = self.get_parameter('on_tx_packet_cb')
        if on_tx_packet is not None:
            if isinstance(msg, AbstractPacket):
                pkt = msg.to_packet()
                pkt = on_tx_packet(pkt)
                if pkt is None:
                    msg = None
                else:
                    msg = msg.from_packet(pkt)

        if msg is not None:
            return super().send_message(message, keep=keep)

        return None


    def on_any_msg(self, message):
        """Handle any incoming message.
        """
        on_rx_packet = self.get_parameter('on_rx_packet_cb')
        if on_rx_packet is not None:

            if isinstance(message, AbstractPacket):
                pkt = message.to_packet()
                if pkt is None:
                    msg = None
                else:
                    pkt = on_rx_packet(pkt)
                    msg = msg.from_packet(pkt)

        if msg is not None:
            return super().on_any_msg(msg)

        return None


class UnixSocketProxy(Thread):
    """Unix socket proxy thread.

    This class provides a convenient way to start a WHAD device proxy
    over a Unix socket, while outputing the socket URL on stdout with
    a set of parameters that will be passed to the next (piped) tool.
    """

    def __init__(self, interface, params, connector=UnixSocketConnector):
        """Create a Unix socket proxy.

        :param WhadDevice   interface   WHAD interface to proxify
        :param dict         params      Dictionnary of params to pass to the next tool
        """
        super().__init__()
        self.__interface = interface
        self.__params = params

        # Remove any previously set interface message filter
        self.__interface.set_queue_filter(None)

        # Connected, switch to a unix socket server
        self.__connector = connector(self.__interface)
        for param in self.__params:
            self.__connector.add_parameter(param, self.__params[param])

    @property
    def interface(self):
        """Get the underlying interface.
        """
        return self.__interface

    @property
    def connector(self):
        """Get the underlying connector.
        """
        return self.__connector

    def stop(self):
        """Stop Unix socket proxy.
        """
        self.__connector.shutdown()

    def run(self):
        """Run device proxy thread.

        This method removes the previous connector associated with the underlying
        device interface (WhadDevice) and connects its own. This connector will
        create a Unix server socket and wait for client connection, and send the
        related URL to stdout along with the provided parameters.

        This method ends when client disconnects.
        """
        self.__connector.serve()



"""
WHAD Bluetooth Low Energy scanner mock device.

This mock emulates a WHAD-compatible hardware connected to the host
and implements the expected behavior of a BLE-oriented WHAD firmware.
"""
import logging
from time import sleep
from random import randint, choice

from typing import List
from whad.device.mock import MockDevice
from whad.hub import ProtocolHub
from whad.hub.discovery import DeviceType, Domain, Capability
from whad.hub.generic.cmdresult import Success, WrongMode, Error
from whad.hub.ble import BDAddress, Commands
from whad.hub.ble.mode import ScanMode, BleStart, BleStop
from whad.hub.ble.pdu import BleAdvPduReceived, AdvType
from whad.hub.ble.sniffing import SniffAdv

from .device import EmulatedDevice

logger = logging.getLogger(__name__)

class DeviceScan(MockDevice):
    """BLE scanning mock device

    This class implements the intended behavior of a BLE-compatible
    hardware interface.
    """

    MODE_SCAN =  0
    MODE_SNIFF = 1

    def __init__(self, devices: List[EmulatedDevice] = None, sniffing: bool = False,
                 nowait: bool = False):
        """Initialize a mock BLE hardware interface that only allow device
        scanning / advertisement sniffing.
        """
        if devices is None:
            logger.warning("No devices passed to scanner")

        # Save sniffing mode status
        self.__sniffing = sniffing
        self.__nowait = nowait

        # By default, no mode enabled and not running
        self.__current_mode = None
        self.__active_scan = False
        self.__bd_filter = None
        self.__running = False
        self.__next_delay = 0

        # Save emulated devices list
        self.__devices = devices

        # Initialize our mock
        super().__init__(
            author="Whad Team",
            url="https://whad.io",
            proto_minver=ProtocolHub.LAST_VERSION,
            version="1.0.0",
            dev_type=DeviceType.VirtualDevice,
            dev_id=b"BleScanMock",
            max_speed=115200,
            capabilities=self.__build_capabilities()
        )

    def __build_capabilities(self) -> dict:
        """Dynamically build the device's capabilities based on its config.
        """
        # By design, we don't capture raw BLE PDUs


        # Default commands
        commands = [ Commands.Start, Commands.Stop ]
        capabilities = Capability.NoRawData

        if self.__sniffing:
            capabilities = capabilities | Capability.Sniff
            commands.append(Commands.SniffAdv)
        else:
            capabilities = capabilities | Capability.Scan
            commands.append(Commands.ScanMode)

        return {
            Domain.BtLE : (
                # We can only scan and sniff with no raw data support
                capabilities,
                commands
            )
        }

    @property
    def started(self) -> bool:
        """Check if mock device is started."""
        return self.__running

    @property
    def stopped(self) -> bool:
        """Check if mock device is stopped."""
        return not self.__running

    def accept(self, bdaddr: BDAddress) -> bool:
        """Apply BD filter if required.
        """
        if self.__bd_filter == b"\xff\xff\xff\xff\xff\xff":
            return True
        return self.__bd_filter == bdaddr.value

    def set_next_delay(self):
        """Generate a random delay between 200 and 1500ms.
        """
        if self.__nowait:
            self.__next_delay = 0
        else:
            self.__next_delay = randint(200, 800)/1000.

    @MockDevice.route(ScanMode)
    def on_scan_mode(self, message: ScanMode):
        """BLE scan mode handler"""
        if not self.__sniffing:
            # Switch to scanning mode
            self.__current_mode = self.MODE_SCAN
            self.__active_scan = message.active
            return Success()
        else:
            return Error()

    @MockDevice.route(SniffAdv)
    def on_sniff_mode(self, message: SniffAdv):
        """BLE Sniffing mode handler"""
        if self.__sniffing:
            self.__current_mode = self.MODE_SNIFF
            self.__bd_filter = message.bd_address
            return Success()
        else:
            return Error()

    @MockDevice.route(BleStart)
    def on_start(self, _: BleStart):
        """Start selected mode"""
        if not self.__running:
            # Generate a random timestamp (in ms) after which we will report
            # one of our emulated devices
            self.set_next_delay()
            # If not running, mark as running and send a success response
            self.__running = True
            return Success()
        else:
            # If already running, return a wrong mode error
            return WrongMode()

    @MockDevice.route(BleStop)
    def on_stop(self, _: BleStop):
        """Stop selected mode."""
        logger.debug("[blescan mock] on_stop()")
        if self.__running:
            self.__running = False
        return Success()

    def on_interface_message(self):
        # Wait for next delay slot
        sleep(self.__next_delay)

        if self.__running:
            report = None

            # Pick a random device from our list
            advertiser = choice(self.__devices)
            if self.__current_mode == self.MODE_SNIFF:
                if self.accept(advertiser.address):
                    # Get advertiser data
                    adv_type, adv_data = advertiser.get_adv_data()

                    # Generate notification message
                    report = BleAdvPduReceived(
                        adv_type=adv_type,
                        rssi=randint(-80, -30),
                        bd_address=advertiser.address.value,
                        adv_data=adv_data,
                        addr_type=advertiser.addr_type
                    )
            else:
                # Generate an advertising notification
                adv_type, adv_data = advertiser.get_adv_data()
                if adv_type!=AdvType.ADV_SCAN_RSP or self.__active_scan:
                    report = BleAdvPduReceived(
                        adv_type=adv_type,
                        rssi=randint(-80, -30),
                        bd_address=advertiser.address.value,
                        adv_data=adv_data,
                        addr_type=advertiser.addr_type
                    )

            if report is not None:
                self.put_message(report)

        # Compute the next delay
        self.set_next_delay()



"""WHAD Phy connectors
"""

from .base import Phy, Endianness, TxPower
from .injector import Injector, InjectionConfiguration
from .lora import LoRa
from .sniffer import Sniffer, SnifferConfiguration

__all__ = [
    "Phy",
    "Endianness",
    "Injector",
    "InjectionConfiguration",
    "LoRa",
    "Sniffer",
    "SnifferConfiguration",
    "TxPower",
]


"""Bluetooth Low Energy validators (mostly crypto-related)
"""
import re
import argparse

from whad.ble.profile.attribute import UUID, InvalidUUIDException as AttrInvalidUUIDException


class InvalidUUIDException(Exception):
    """Base class for Invalid HandleException -- The attribute handle given was
    not valid on this server."""

    def __init__(self, uuid):
        self.code = 0x01
        self.name = "Invalid UUID"
        self.description = (
            f"The supplied UUID ({uuid}) is not a valid 16-bit or 128-bit UUID. "
            "(ex: 2a1b or 000AA000-0BB0-10C0-80A0-00805F9B34FB"
        )
        

# Validators for various BLESuite input parameters supplied by the user. The use
# of these validators is mostly limited to import/export and CLI features.


class InvalidBDADDRException(Exception):
    """Invalid Bluetooth Device address.
    """

    def __init__(self, address):
        self.code = 0x02
        self.name = "Invalid BD_ADDR"
        self.description = (
            f"The supplied BD_ADDR ({address}) is not valid. Use format 00:11:22:33:44:55"
        )


class InvalidAddressTypeByName(Exception):
    """Invalid Bluetooth address type.

    Bluetooth device address is defined by the combination of TxAdd/RxAdd,
    and must match one of these two types: random or public.
    """

    def __init__(self, address_type):
        self.code = 0x02
        self.name = "Invalid address type name"
        self.description = (
            f"The supplied address type name ({address_type}) is not valid. "
            f"Options: ['public', 'random']"
        )


class InvalidATTHandle(Exception):
    """Invalid ATT protocol handle.
    """

    def __init__(self, handle):
        self.code = 0x03
        self.name = "Invalid ATT handle"
        self.description = (
            f"The supplied attribute handle ({handle}) is not valid. "
            f"The integer value must be >= 1 and <= 0xffff"
        )


class InvalidATTSecurityMode(Exception):
    """Invalid security mode.
    """

    def __init__(self, mode, level):
        self.code = 0x04
        self.name = "Invalid ATT security mode"
        self.description = (
            f"The supplied attribute security mode ({mode}, {level}) is not valid. "
            "Options: (Security Mode, Security Level) [(0,0) -- No Access, (1,1) -- "
            "Open, (1,2) -- Requires encryption and does not require authenticated pairing,"
            " (1,3) -- Requires encryption and requires authenticated pairing, (1,4)"
            " -- Requires encryption with Secure Connections pairing, (2,1)"
            " -- Data signing and does not require authenticated pairing, (2,2)"
            " -- Data signing and requires authenticated pairing]"
        )


class InvalidATTProperty(Exception):
    """Invalid attribute property.
    """

    def __init__(self, att_property):
        self.code = 0x05
        self.name = "Invalid ATT property"
        self.description = (
            f"The supplied attribute property ({att_property}) is not valid. "
            "Options: ['read', 'write']"
        )


class InvalidGATTProperty(Exception):
    """Invalid GATT property.
    """

    def __init__(self, gatt_property):
        self.code = 0x06
        self.name = "Invalid GATT property"
        self.description = (
            f"The supplied GATT property ({gatt_property}) is not valid. "
            "Options: ['broadcast', 'read', 'write', 'notify', 'indicate', "
            "'authenticated signed writes', 'extended properties']"
        )


class InvalidSMLTK(Exception):
    """Invalid Long-term key (LTK).
    """

    def __init__(self, ltk):
        self.code = 0x07
        self.name = "Invalid SM LTK"
        self.description = (
            f"The supplied SM LTK ({ltk}) is not valid. Expects hex string (ie 'AB7F2A...')"
        )


class InvalidSMRandom(Exception):
    """Invalid Random.
    """

    def __init__(self, rand):
        self.code = 0x08
        self.name = "Invalid SM Random"
        self.description = (
            f"The supplied SM Random ({rand}) is not valid. Expects hex string (ie 'AB7F2A...')"
        )


class InvalidSMIRK(Exception):
    """Invalid IRK.
    """

    def __init__(self, irk):
        self.code = 0x09
        self.name = "Invalid SM IRK"
        self.description = (
            f"The supplied SM IRK ({irk}) is not valid. Expects hex string (ie 'AB7F2A...')"
        )


class InvalidSMCSRK(Exception):
    """Invalid CSRK.
    """

    def __init__(self, csrk):
        self.code = 0x0a
        self.name = "Invalid SM CSRK"
        self.description = (
            f"The supplied SM CSRK ({csrk}) is not valid. Expects hex string (ie 'AB7F2A...')"
        )


def validate_bluetooth_address_cli(address):
    """
    Validates BT address string

    :param address: BT Address
    :type: str
    :return: address
    :raises: ArgumentTypeError
    """
    if address is not None:
        match = re.search(("^[0-9A-Fa-f]{2}:[0-9A-Fa-f]{2}:[0-9A-Fa-f]{2}:[0-9A-Fa-f]{2}"
                           ":[0-9A-Fa-f]{2}:[0-9A-Fa-f]{2}$"), address)
        if match is not None and match.group(0) is not None:
            return address

    raise argparse.ArgumentTypeError(f"{address} is an invalid Bluetooth address (BD_ADDR)")


def validate_bluetooth_address(address):
    """
    Validates BT address string

    :param address: BT Address
    :type: str
    :return: address
    :raises: InvalidBDADDRException
    """
    if address is not None:
        match = re.search(("^[0-9A-Fa-f]{2}:[0-9A-Fa-f]{2}:[0-9A-Fa-f]{2}"
                           ":[0-9A-Fa-f]{2}:[0-9A-Fa-f]{2}:[0-9A-Fa-f]{2}$"),
                          address)
        if match is not None and match.group(0) is not None:
            return address

    raise InvalidBDADDRException(address)


def validate_attribute_uuid(uuid):
    """Validate attribute UUID.

    :param uuid: Attribute UUID
    :type uuid: UUID
    :return: Validated UUID
    :raises: InvalidUUIDException
    """
    if uuid is None:
        raise InvalidUUIDException(uuid)
    # When attribute UUID is read from JSON, it can be encoded as unicode,
    # which will break the UUID class
    try:
        UUID(uuid)
    except AttrInvalidUUIDException as uuid_err:
        raise InvalidUUIDException(uuid) from uuid_err
    except TypeError as type_err:
        raise InvalidUUIDException(uuid) from type_err
    return uuid


def validate_address_type_name(address_type_name: str) -> str:
    """Validate address_type_name (must be 'random' or 'public').

    :param address_type_name: Address type name
    :type address_type_name: str
    :return: Validated address type name
    :rtype: str
    :raises: InvalidAddressTypeByName
    """
    if address_type_name is None:
        return InvalidAddressTypeByName(address_type_name)
    address_type_name = address_type_name.lower()
    if address_type_name in ("public", "random"):
        return address_type_name

    raise InvalidAddressTypeByName(address_type_name)


def validate_int_att_handle(handle: int) -> int:
    """Validate ATT handle value.

    :param handle: Attribute handle to validate
    :type handle: int
    :return: Validated attribute handle
    :rtype: int
    :raises: InvalidATTHandle
    """
    if handle is None or handle < 0x01 or handle > 0xffff:
        raise InvalidATTHandle
    return handle


def validate_att_security_mode(mode: int, level: int) -> tuple[int, int]:
    """Validate ATT security mode.

    :param mode: Security mode to validate
    :type mode: int
    :param level: Security level to validate
    :type level: int
    :return: Validated security mode and level
    :rtype: tuple
    :raises: InvalidATTSecurityMode
    """
    supported_modes = [
        (0, 0), (1, 1), (1, 2), (1, 3), (1, 4), (2, 1), (2, 2)
    ]

    if (mode, level) not in supported_modes:
        raise InvalidATTSecurityMode(mode, level)
    return mode, level


def validate_att_property(prop: str) -> str:
    """Validate ATT property.

    :param prop: Attribute property to validate
    :type prop: str
    :return: Validated property
    :rtype: str
    :raises: InvalidATTProperty
    """
    if prop.lower() not in ("read", "write"):
        raise InvalidATTProperty(prop)
    return prop


def validate_gatt_property(prop: str) -> str:
    """Validate GATT attribute property.

    :param prop: GATT property
    :type prop: str
    :return: Validated GATT property
    :rtype: str
    :raises: InvalidGATTProperty
    """
    prop = prop.lower()
    valid_properties = ['broadcast', 'read', 'write', 'notify', 'indicate',
                        'authenticated signed writes', 'extended properties',
                        'write without response']
    if prop not in valid_properties:
        raise InvalidGATTProperty(prop)
    return prop


def validate_ltk(ltk: str) -> str:
    """Validate LTK

    :param ltk: Long-term key in hex
    :type ltk: str
    :return: Validated LTK
    :rtype: str
    :raises: InvalidSMLTK
    """
    ltk = ltk.lower()
    try:
        bytes.fromhex(ltk)
    except ValueError as err_type:
        raise InvalidSMLTK(ltk) from err_type
    return ltk


def validate_irk(irk: str) -> str:
    """Validate IRK

    :param irk: IRK to validate (in hex)
    :type irk: str
    :return: Validated IRK
    :rtype: str
    :raises: InvalidSMIRK
    """
    if irk is None:
        irk = "00" * 16
    irk = irk.lower()
    try:
        bytes.fromhex(irk)
    except ValueError as err_type:
        raise InvalidSMIRK(irk) from err_type
    return irk


def validate_csrk(csrk: str) -> str:
    """Validate CSRK

    :param csrk: CSRK to validate (in hex)
    :type csrk: str
    :return: Validated CSRK
    :rtype: str
    :raises: InvalidSMCSRK
    """
    if csrk is None:
        csrk = "00" * 16
    csrk = csrk.lower()
    try:
        bytes.fromhex(csrk)
    except ValueError as err_type:
        raise InvalidSMCSRK(csrk) from err_type
    return csrk


def validate_rand(rand):
    """Validate Rand value

    :param ramd: Rand value to validate (in hex)
    :type rand: str
    :return: Validated Rand
    :rtype: str
    :raises: InvalidSMRandom
    """
    rand = rand.lower()
    try:
        bytes.fromhex(rand)
    except ValueError as err_type:
        raise InvalidSMRandom(rand) from err_type
    return rand



"""WHAD Protocol BLE jamming messages abstraction layer.
"""
from whad.protocol.whad_pb2 import Message
from whad.protocol.ble.ble_pb2 import JamAdvCmd
from whad.hub.message import pb_bind, PbFieldInt, PbFieldBytes, PbMessageWrapper
from whad.hub.ble import BleDomain

@pb_bind(BleDomain, "jam_adv", 1)
class JamAdv(PbMessageWrapper):
    """BLE advertisement jamming message class
    """
    
    def __init__(self, message: Message = None):
        super().__init__(message=message)
        self.message.ble.jam_adv.CopyFrom(JamAdvCmd())

@pb_bind(BleDomain, "jam_adv_chan", 1)
class JamAdvChan(PbMessageWrapper):
    """BLE advertisement channel jamming message class
    """
    channel = PbFieldInt("ble.jam_adv_chan.channel")


@pb_bind(BleDomain, "jam_conn", 1)
class JamConn(PbMessageWrapper):
    """BLE advertisement channel jamming message class
    """
    access_address = PbFieldInt("ble.jam_conn.access_address")

@pb_bind(BleDomain, "reactive_jam", 1)
class ReactiveJam(PbMessageWrapper):
    """BLE reactive jamming message class
    """
    channel = PbFieldInt("ble.reactive_jam.channel")
    pattern = PbFieldBytes("ble.reactive_jam.pattern")
    position = PbFieldInt("ble.reactive_jam.position")
